import logging
import seagrass._typing as t
from collections import Counter
from seagrass.base import ProtoHook
from ._utils import HookContext


class CounterHook(ProtoHook[None]):
    """A Seagrass hook that counts the number of times an event occurs.

    **Examples:**

    .. testsetup:: counterhook-examples

       from seagrass import Auditor
       from seagrass._docs import configure_logging

       configure_logging()
       auditor = Auditor()

    .. doctest:: counterhook-examples

       >>> from seagrass.hooks import CounterHook

       >>> hook = CounterHook()

       >>> event_a = auditor.create_event("event_a", hooks=[hook])

       >>> event_b = auditor.create_event("event_b", hooks=[hook])

       >>> with auditor.start_auditing(log_results=True):
       ...     for _ in range(15):
       ...         auditor.raise_event("event_a")
       ...     for _ in range(8):
       ...         auditor.raise_event("event_b")
       {"message": "CounterHook results", "seagrass": {"event": null, "hook": "CounterHook", \
"hook_ctx": {"event": "event_a", "count": 15}}, "level": "INFO"}
       {"message": "CounterHook results", "seagrass": {"event": null, "hook": "CounterHook", \
"hook_ctx": {"event": "event_b", "count": 8}}, "level": "INFO"}
    """

    event_counter: t.Counter[str]

    def __init__(self) -> None:
        self.event_counter = Counter()

    def prehook(
        self, event_name: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> None:
        self.event_counter[event_name] += 1

    def reset(self) -> None:
        self.event_counter.clear()

    def log_results(self, logger: logging.Logger) -> None:
        if len(self.event_counter) == 0:
            with HookContext("CounterHook", {}):
                logger.warning("no events recorded by counter")
            return

        for event in sorted(self.event_counter):
            ctx = {"event": event, "count": self.event_counter[event]}
            with HookContext("CounterHook", ctx):
                logger.info("CounterHook results")



# Basic errors that can be thrown while using Seagrass


class SeagrassError(Exception):
    """A generic error for the Seagrass library."""


class EventNotFoundError(SeagrassError):
    """Raised when we try to reference an auditing event that does not currently
    exist."""

    event_name: str

    def __init__(self, event_name: str) -> None:
        """Create a new EventNotFoundError.

        :param str event_name: the name of the event that could not be found.
        """
        self.event_name = event_name
        msg = f"Audit event not found: {event_name}"
        super().__init__(msg)



# Tests for ProfilerHook

import time
import unittest
from test.utils import HookTestCaseMixin, req_python_version
from seagrass.base import LogResultsHook, ResettableHook, CleanupHook
from seagrass.hooks import ProfilerHook


class ProfilerHookTestCase(HookTestCaseMixin, unittest.TestCase):

    check_interfaces = (LogResultsHook, ResettableHook, CleanupHook)

    @staticmethod
    def hook_gen():
        return ProfilerHook(sort_keys="cumtime", restrictions=0.1)

    # Test only works for Python >= 3.9 due to the use of StatsProfile
    @req_python_version(min=(3, 9))
    def test_hook_function(self):
        # Note: could just as easily use auditor.audit("test.sleep", time.sleep, ...)
        # here but the name of time.sleep is slightly mangled in the resulting
        # StatsProfile that we generate, which complicates testing.
        @self.auditor.audit("test.sleep", hooks=[self.hook])
        def ausleep(*args):
            time.sleep(*args)

        self.assertEqual(self.hook.get_stats(), None)

        with self.auditor.start_auditing():
            for _ in range(10):
                ausleep(0.001)

        # Get profiler information for ausleep
        stats_profile = self.hook.get_stats().get_stats_profile()
        ausleep_profile = stats_profile.func_profiles["ausleep"]
        self.assertEqual(ausleep_profile.ncalls, "10")

        # Profiler information should be reset after hook.reset() is called
        self.hook.reset()
        self.assertEqual(self.hook.get_stats(), None)

        with self.auditor.start_auditing():
            ausleep(0.01)

        stats_profile = self.hook.get_stats().get_stats_profile()
        ausleep_profile = stats_profile.func_profiles["ausleep"]
        self.assertEqual(ausleep_profile.ncalls, "1")
        self.assertAlmostEqual(ausleep_profile.cumtime, 0.01, delta=0.005)

    # Test only works for Python >= 3.9 due to the use of StatsProfile
    @req_python_version(min=(3, 9))
    def test_hook_nested_functions(self):
        # Ensure that the profiler collects information about nested events
        @self.auditor.audit("test.foo", hooks=[self.hook])
        def foo(*args):
            time.sleep(*args)

        @self.auditor.audit("test.bar", hooks=[self.hook])
        def bar(*args):
            foo(0)
            foo(*args)

        @self.auditor.audit("test.baz", hooks=[self.hook])
        def baz(*args):
            bar(0)
            bar(*args)

        with self.auditor.start_auditing():
            baz(0.05)

        stats_profile = self.hook.get_stats().get_stats_profile()
        foo_profile = stats_profile.func_profiles["foo"]
        bar_profile = stats_profile.func_profiles["bar"]
        baz_profile = stats_profile.func_profiles["baz"]

        self.assertTrue(foo_profile.cumtime >= 0.05)
        self.assertTrue(bar_profile.cumtime >= 0.05)
        self.assertTrue(baz_profile.cumtime >= 0.05)

        self.assertEqual(foo_profile.ncalls, "4")
        self.assertEqual(bar_profile.ncalls, "2")
        self.assertEqual(baz_profile.ncalls, "1")



import functools
import logging
import seagrass._typing as t
from contextlib import contextmanager
from contextvars import ContextVar
from seagrass._logging import SeagrassLogFilter
from seagrass.base import LogResultsHook, ProtoHook, ResettableHook
from seagrass.errors import EventNotFoundError
from seagrass.events import Event, AsyncEvent, SyncEvent

# The name of the default logger used by Seagrass
DEFAULT_LOGGER_NAME: t.Final[str] = "seagrass"

# A context variable that keeps track of the current Auditor, if
# one is being executed.
_current_auditor: ContextVar["Auditor"] = ContextVar("seagrass_auditor")

F = t.TypeVar("F", bound=t.Callable)
T = t.TypeVar("T")


def _empty_event_func(*args, **kwargs) -> None:
    """A function used to define empty events. This function can take an arbitrary combination
    of parameters, but internally it does nothing."""


class Auditor:
    """
    An auditing instance that allows you to dynamically audit and profile
    code.
    """

    events: t.Dict[str, Event]
    event_wrappers: t.Dict[str, t.Callable]
    hooks: t.Set[ProtoHook]
    logger_name: str
    __logger_filter: SeagrassLogFilter
    __enabled: bool = False

    def __init__(self, logger: str = DEFAULT_LOGGER_NAME) -> None:
        """Create a new Auditor instance.

        :param Union[str,logging.Logger] logger: The logger that this auditor should use. When set
            to a string the auditor uses the logger returned by ``logging.getLogger(logger)``.
        """
        self.logger_name = logger
        self.events = dict()
        self.event_wrappers = dict()
        self.hooks = set()
        self.__logger_filter = SeagrassLogFilter()
        self.reset_filter()

    @property
    def logger(self) -> logging.Logger:
        logger = logging.getLogger(self.logger_name)
        logger.addFilter(self.__logger_filter)
        return logger

    @property
    def enabled(self) -> bool:
        """Return whether or not the auditor is enabled.

        :type: bool
        """
        return self.__enabled

    @property
    def event_filter(self) -> t.Callable[[str], bool]:
        """A filter over the events being audited.

        :param EventFilter filter: the filter that should be added to the auditor. The filter
            should satisfy the :py:class:`seagrass.events.EventFilter` interface -- that is, they
            should be functions that take a single string and return a bool.
        :raises TypeError: if the filters don't satisfy the
            :py:class:`~seagrass.events.EventFilter` interface.

        **Examples:**

        .. testsetup:: filter-events-doctests

            from seagrass import Auditor
            from seagrass._docs import configure_logging
            configure_logging()
            auditor = Auditor()

        .. doctest:: filter-events-doctests

            >>> from seagrass.hooks import LoggingHook

            >>> hook = LoggingHook(prehook_msg = lambda event, *args: f"{event} called")

            >>> _ = auditor.create_event("example.foo", hooks=[hook])

            >>> with auditor.start_auditing():
            ...     auditor.raise_event("example.foo")
            {"message": "example.foo called", "seagrass": {"event": "example.foo"}, "level": "DEBUG"}

            >>> auditor.event_filter = lambda event: not event.startswith("example.")

            >>> with auditor.start_auditing():
            ...     auditor.raise_event("example.foo")
        """
        return self.__event_filter

    @event_filter.setter
    def event_filter(self, filter: t.Callable[[str], bool]):
        self.__event_filter = filter

    def reset_filter(self) -> None:
        """Reset the event filter used by the Auditor."""
        self.event_filter = lambda event: True

    def toggle_auditing(self, mode: bool) -> None:
        """Enable or disable auditing.

        :param bool mode: When set to ``True``, auditing is enabled; when set to ``False``,
            auditing is disabled.
        """
        self.__enabled = mode

    @contextmanager
    def start_auditing(
        self,
        filter: t.Optional[t.Callable[[str], bool]] = None,
        reset_hooks: bool = False,
        log_results: bool = False,
    ) -> t.Iterator[None]:
        """Create a new context within which the auditor is enabled. You can replicate this
        functionality by calling :py:meth:`toggle_auditing`, e.g.

        .. testsetup::

            from seagrass import Auditor
            auditor = Auditor()

        .. testcode::

            try:
                auditor.toggle_auditing(True)
                # Put code under audit here
                ...
            finally:
                auditor.toggle_auditing(False)

        However, using ``with auditor.start_auditing()`` in place of ``auditor.toggle_auditing`` has
        some additional benefits too, e.g. it allows you to access the logger for the most recent
        auditing context using ``seagrass.get_audit_logger``.

        :param Callable[[str],bool] filter: a filter to apply on which events should be audited.
        :param bool log_results: Log hooks results with :py:meth:`log_results` before exiting
            the auditing context.
        :param bool reset_hooks: Reset hooks with :py:meth:`reset`: before exiting the
            auditing context.
        """
        try:
            auditor_token = _current_auditor.set(self)

            if filter is not None:
                old_filter = self.event_filter
                self.event_filter = filter

            self.toggle_auditing(True)
            yield None

        finally:
            self.toggle_auditing(False)
            _current_auditor.reset(auditor_token)

            if filter is not None:
                self.event_filter = old_filter

            if log_results:
                self.log_results()
            if reset_hooks:
                self.reset_hooks()

    # Overload the audit function so that it can be called either as a decorator or as
    # a regular function.

    @t.overload
    def audit(
        self,
        event_name: t.Union[str, t.Callable[[F], str]],
        func: None = None,
        hooks: t.Optional[t.List[ProtoHook]] = None,
        use_async: bool = False,
        **kwargs,
    ) -> t.AuditDecorator:
        ...  # pragma: no cover

    @t.overload
    def audit(
        self,
        event_name: t.Union[str, t.Callable[[F], str]],
        func: F,
        hooks: t.Optional[t.List[ProtoHook]] = None,
        use_async: bool = False,
        **kwargs,
    ) -> F:
        ...  # pragma: no cover

    def audit(
        self,
        event_name: t.Union[str, t.Callable[[F], str]],
        func: t.Optional[F] = None,
        hooks: t.Optional[t.List[ProtoHook]] = None,
        use_async: bool = False,
        **kwargs,
    ) -> t.Union[t.AuditDecorator, F]:
        """Wrap a function with a new auditing event. You can call ``audit`` either as a function
        decorator or as a regular method of :py:class:`Auditor`.

        :param Union[str,Callable[[F],str]] event_name: the name of the new event, which must be
            unique. This parameter can either be a string, or it can be a function that takes the
            audited function and creates a string from it, e.g.
            ``event_name = lambda func: f"event.{func.__name__}"``.
        :param bool use_async: wrap the function in an asynchronous event. This should be set to
            ``True`` if ``func`` is ``async``.
        :param Optional[Callable] func: the function that should be wrapped in a new event.
        :param Optional[List[ProtoHook]] hooks: a list of hooks to call whenever the new event is
            triggered.
        :param kwargs: keyword arguments to pass on to ``Event.__init__``.

        **Examples:** create an event over the function ``json.dumps`` using ``wrap``:

        .. testsetup::

            from seagrass import Auditor
            auditor = Auditor()

        .. doctest::

            >>> import json
            >>> from seagrass.hooks import CounterHook
            >>> hook = CounterHook()
            >>> audumps = auditor.audit("audit.json.dumps", json.dumps, hooks=[hook])
            >>> setattr(json, "dumps", audumps)
            >>> hook.event_counter["audit.json.dumps"]
            0
            >>> with auditor.start_auditing():
            ...     json.dumps({"a": 1, "b": 2})
            '{"a": 1, "b": 2}'
            >>> hook.event_counter["audit.json.dumps"]
            1

        Here is another example where we call ``auditor.audit`` as a decorator for a function
        ``add``:

        .. testcode::

            from seagrass import Auditor
            from seagrass.hooks import CounterHook
            auditor = Auditor()

            @auditor.audit("event.add", hooks=[CounterHook()])
            def add(x, y):
                return x + y

        """

        # If func is None, we assume that audit() was called as a function decorator, and
        # we return another decorator that can be called around the function.
        if func is None:

            def decorator(func: F) -> F:
                return self.audit(
                    event_name, func, hooks=hooks, use_async=use_async, **kwargs
                )

            return t.cast(t.AuditDecorator, decorator)

        if callable(event_name):
            event_name = event_name(func)

        if event_name in self.events:
            raise ValueError(
                f"An event with the name '{event_name}' has already been defined"
            )

        hooks = [] if hooks is None else hooks

        # Add hooks to the Auditor's `hooks` set
        for hook in hooks:
            self.hooks.add(hook)

        if use_async:
            new_event: t.Union[AsyncEvent, SyncEvent] = AsyncEvent(
                func, event_name, hooks=hooks, **kwargs
            )
        else:
            new_event = SyncEvent(func, event_name, hooks=hooks, **kwargs)

        self.events[event_name] = new_event

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not self.enabled or not self.event_filter(event_name):
                return new_event.func(*args, **kwargs)
            else:
                return new_event(*args, **kwargs)

        # Add an __event_name__ attribute to the function so that users can easily look up
        # the name of the event. This is especially useful for events that are auto-named.
        audit_func = t.cast(t.AuditedFunc[F], wrapper)
        audit_func.__event_name__ = event_name

        self.event_wrappers[event_name] = audit_func
        return audit_func

    def async_audit(
        self,
        event_name: t.Union[str, t.Callable[[F], str]],
        func: t.Optional[F] = None,
        hooks: t.Optional[t.List[ProtoHook]] = None,
        **kwargs,
    ) -> t.Union[t.AuditDecorator, F]:
        """Call :py:meth:`audit` with `use_async=True`. See the documentation for :py:meth:`audit`
        for more information.
        """
        return self.audit(event_name, func, use_async=True, hooks=hooks, **kwargs)

    def create_event(self, event_name: str, **kwargs) -> t.Callable[..., None]:
        """Create a new "empty" event. When this event is executed, it runs any hooks that are
        associated with the event, but the function wrapped by the event itself does nothing.

        :param str event_name: the name of the event that should be created. Event names should
            be unique.
        :param kwargs: keyword arguments. The keyword arguments for this function are the same
            as those for :py:meth:`wrap`.
        :return: returns a wrapper function around the event that was created.

        **Example:**

        .. doctest::

            >>> from seagrass import Auditor

            >>> from seagrass.hooks import CounterHook

            >>> auditor = Auditor()

            >>> hook = CounterHook()

            >>> wrapper = auditor.create_event("my_signal", hooks=[hook])

            >>> hook.event_counter["my_signal"]
            0

            >>> with auditor.start_auditing():
            ...     auditor.raise_event("my_signal")

            >>> hook.event_counter["my_signal"]
            1
        """
        return self.audit(event_name, _empty_event_func, **kwargs)

    def reset_hooks(self) -> None:
        """Reset all of the hooks used by this Auditor."""
        for hook in self.hooks:
            if isinstance(hook, ResettableHook):
                hook.reset()

    def raise_event(self, event_name: str, *args, **kwargs) -> t.Any:
        """Trigger an audit event using the input arguments and keyword arguments.

        :param str event_name: the name of the event to be raised.
        :param args: arguments to pass to the event.
        :param kwargs: keyword arguments to pass to the event.
        :return: returns the output of the event that was called.
        :rtype: Any
        :raises seagrass.errors.EventNotFoundError: if the auditor can't find the event with the
            provided name.
        """

        wrapper = self.event_wrappers.get(event_name)
        if wrapper is not None:
            return wrapper(*args, **kwargs)
        else:
            raise EventNotFoundError(event_name)

    def add_hooks(self, event_name: str, *hooks: ProtoHook) -> None:
        """Add new hooks to an auditing event.

        :param str event_name: the name of the event to add the hooks to.
        :param ProtoHook hooks: the hooks that should be added to the event.
        :raises seagrass.errors.EventNotFoundError: if the auditor can't find the event with the
            provided name.
        """
        event = self.events.get(event_name)
        if event is not None:
            event.add_hooks(*hooks)
        else:
            raise EventNotFoundError(event_name)

    def toggle_event(self, event_name: str, enabled: bool) -> None:
        """Enables or disables an auditing event.

        :param str event_name: the name of the event to toggle.
        :param bool enabled: whether to enable or disabled the event.

        **Example:**

        .. testsetup::

            from seagrass import Auditor
            auditor = Auditor()

        .. doctest::

            >>> from seagrass.hooks import CounterHook
            >>> hook = CounterHook()
            >>> @auditor.audit("event.say_hello", hooks=[hook])
            ... def say_hello(name):
            ...     return f"Hello, {name}!"
            >>> hook.event_counter["event.say_hello"]
            0
            >>> with auditor.start_auditing():
            ...     say_hello("Alice")
            'Hello, Alice!'
            >>> hook.event_counter["event.say_hello"]
            1
            >>> # Disable the "event.say_hello" event
            >>> auditor.toggle_event("event.say_hello", False)
            >>> with auditor.start_auditing():
            ...     # Since event.say_hello is disabled, the following call to
            ...     # say_hello will not contribute to its event counter.
            ...     say_hello("Bob")
            'Hello, Bob!'
            >>> hook.event_counter["event.say_hello"]
            1

        """
        self.events[event_name].enabled = enabled

    def log_results(self) -> None:
        """Log results stored by hooks by calling `log_results` on all
        :py:class:`~seagrass.base.LogResultsHook` hooks."""
        for hook in self.hooks:
            if isinstance(hook, LogResultsHook):
                hook.log_results(self.logger)


@t.overload
def get_auditor() -> Auditor:
    ...  # pragma: no cover


@t.overload
def get_auditor(default: t.Missing) -> Auditor:
    ...  # pragma: no cover


@t.overload
def get_auditor(default: T) -> t.Union[Auditor, T]:
    ...  # pragma: no cover


def get_auditor(default: t.Maybe[T] = t.MISSING) -> t.Union[Auditor, T]:
    """Get the current auditor, if in an active auditing context.

    This function only works in auditing contexts created by :py:meth:`Auditor.audit`; it will
    be unable to get the auditor for the current auditing context if you use
    :py:meth:`Auditor.toggle_auditing`.

    :return: the auditor for the current auditing context, or ``default`` if no auditing
        context has been created.
    :rtype: t.Union[logging.Logger,T]
    :raises LookupError: if this function is called outside of an auditing context (and no
        ``default`` is provided).
    """

    if isinstance(default, t.Missing):
        return _current_auditor.get()
    else:
        return _current_auditor.get(default)


@t.overload
def get_audit_logger() -> logging.Logger:
    ...  # pragma: no cover


@t.overload
def get_audit_logger(default: t.Missing) -> logging.Logger:
    ...  # pragma: no cover


@t.overload
def get_audit_logger(default: T) -> t.Union[logging.Logger, T]:
    ...  # pragma: no cover


def get_audit_logger(default: t.Maybe[T] = t.MISSING) -> t.Union[logging.Logger, T]:
    """Get the logger belonging to the auditor in the current auditing context.

    This function only works in auditing contexts created by :py:meth:`Auditor.audit`; it will
    be unable to get the logger for the current auditing context if you use
    :py:meth:`Auditor.toggle_auditing`.

    :return: the logger for the most recent auditing context, or ``default`` if no auditing
        context has been created.
    :rtype: t.Union[logging.Logger,T]
    :raises LookupError: if this function is called outside of an auditing context (and no
        ``default`` is provided).
    """

    if isinstance(default, t.Missing):
        return get_auditor().logger

    auditor = get_auditor(None)
    if auditor is None:
        return default

    return auditor.logger



# Testing utilities and base classes for testing Seagrass

import asyncio
import datetime
import logging
import logging.config
import sys
import seagrass._typing as t
from functools import wraps
from io import StringIO
from pythonjsonlogger import jsonlogger
from seagrass import Auditor
from seagrass.base import ProtoHook


class LogFormatter(jsonlogger.JsonFormatter):   # type: ignore
    def add_fields(self, log_record, record, message_dict):
        super(LogFormatter, self).add_fields(log_record, record, message_dict)

        timestamp = datetime.datetime.fromtimestamp(record.created)
        log_record["timestamp"] = timestamp.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        log_record["level"] = record.levelname


def async_test(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(func(*args, **kwargs))

    return wrapper


class SeagrassTestCaseMixin:

    logging_output: StringIO
    auditor: Auditor

    # Whether to emit logs that are >= WARNING to stdout
    log_warnings_to_stdout: bool = True

    @property
    def logger(self) -> logging.Logger:
        return logging.getLogger(self.__logger_name)

    def setUp(self) -> None:
        # Set up logging configuration
        self.logging_output = StringIO()
        self.__logger_name = "test.seagrass"

        handlers = ["default"]
        if self.log_warnings_to_stdout:
            handlers.append("warnings")

        formatter = LogFormatter()

        self.log_buffer_handler = logging.StreamHandler(stream=self.logging_output)
        self.log_buffer_handler.setLevel(logging.DEBUG)
        self.log_buffer_handler.setFormatter(formatter)

        self.log_stdout_handler = logging.StreamHandler()
        self.log_stdout_handler.setLevel(logging.WARNING)
        self.log_stdout_handler.setFormatter(formatter)

        self.logger.addHandler(self.log_buffer_handler)
        if self.log_warnings_to_stdout:
            self.logger.addHandler(self.log_stdout_handler)
        self.logger.setLevel(logging.DEBUG)

        # Create a new auditor instance with the logger we just
        # set up
        self.auditor = Auditor(logger=self.__logger_name)

    def tearDown(self):
        self.logger.removeHandler(self.log_buffer_handler)
        if self.log_warnings_to_stdout:
            self.logger.removeHandler(self.log_stdout_handler)


class HookTestCaseMixin(SeagrassTestCaseMixin):
    """A base testing class for auditor hooks."""

    hook: ProtoHook
    hook_gen: t.Optional[t.Callable[[], ProtoHook]] = None

    # A list of all of the interfaces that the hook is expected
    # to satisfy.
    check_interfaces: t.Optional[t.Tuple[t.Type, ...]] = None

    @property
    def hook_name(self) -> str:
        return self.hook.__class__.__name__

    def setUp(self):
        super().setUp()

        if self.hook_gen is not None:
            self.hook = self.hook_gen()

    def test_hook_satisfies_interfaces(self):
        CheckableProtoHook = t.runtime_checkable(ProtoHook)
        self.assertIsInstance(
            self.hook,
            CheckableProtoHook,
            f"{self.hook_name} does not satisfy the ProtoHook interface",
        )

        interfaces = [] if self.check_interfaces is None else self.check_interfaces

        for interface in interfaces:
            self.assertIsInstance(
                self.hook,
                interface,
                f"{self.hook_name} does not satisfy the {interface.__name__} interface",
            )


F = t.Callable[..., None]


def req_python_version(
    min: t.Optional[t.Tuple[int, int]] = None, max: t.Optional[t.Tuple[int, int]] = None
) -> t.Callable[[F], F]:
    """Create a wrapper around a unit test to enforce a minimum
    or maximum Python version."""

    assert min is not None or max is not None
    version = sys.version_info

    err_msg = f"Test skipped: disabled for Python version = {version.major}.{version.minor}.{version.micro}, required"
    bad_version = False
    if (min is not None and max is not None) and (version < min or max < version):
        err_msg = f"{err_msg} {min[0]}.{min[1]} <= version <= {max[0]}.{max[1]}"
        bad_version = True
    elif (min is not None and max is None) and (version < min):
        err_msg = f"{err_msg} version >= {min[0]}.{min[1]}"
        bad_version = True
    elif (min is None and max is not None) and (version >= max):
        err_msg = f"{err_msg} version < {max[0]}.{max[1]}"
        bad_version = True

    if not bad_version:

        def decorator(test_func: F) -> F:
            return test_func

        return decorator
    else:

        def decorator(test_func: F) -> F:
            @wraps(test_func)
            def wrapper(self, *args, **kwargs):
                self.skipTest(err_msg)

            return wrapper

        return decorator



# Tests for hook satisfying the CleanupHook interface.

import unittest
from seagrass.base import ProtoHook, CleanupHook, ResettableHook
from test.utils import SeagrassTestCaseMixin


class CleanupHookTestCase(SeagrassTestCaseMixin, unittest.TestCase):
    """Tests to check that the cleanup stage of hooks that satisfy the CleanupHook interface
    is unconditionally executed."""

    class _BaseTestHook(ProtoHook[None]):
        def __init__(self):
            self.reset()

        def prehook(self, *args):
            pass

        def reset(self):
            self.counter = 0
            self.exception = None

    class _HookA(_BaseTestHook):
        # Regular hook with no cleanup stage
        def posthook(self, *args):
            self.counter += 1

    class _HookB(_BaseTestHook):
        # A hook that satisfies the CleanupHook interface
        def posthook(self, *args):
            pass

        def cleanup(self, event, context, exc):
            self.counter += 1
            self.exception = exc

    class _HookC(_BaseTestHook):
        # A hook that satisfies the CleanupHook interface, but which also raises an
        # exception in its posthook.
        def posthook(self, *args):
            assert False

        def cleanup(self, event, context, exc):
            self.counter += 1
            self.exception = exc

    class _HookD(_BaseTestHook):
        # A hook that satisfies the CleanupHook interface, but which also raises an
        # exception in its prehook.
        def prehook(self, *args):
            assert False

        def posthook(self, *args):
            pass

        def cleanup(self, event, context, exc):
            self.counter += 1
            self.exception = exc

    def setUp(self):
        super().setUp()
        self.hook_a = self._HookA()
        self.hook_b = self._HookB()
        self.hook_c = self._HookC()
        self.hook_d = self._HookD()
        self.hooks = (self.hook_a, self.hook_b, self.hook_c, self.hook_d)
        self.ex = None

    def test_hooks_satisfy_interfaces(self):
        # All of the hooks, except for _HookA, should satisfy the CleanupHook interface.
        self.assertNotIsInstance(self.hooks[0], CleanupHook)
        for hook in self.hooks[1:]:
            self.assertIsInstance(hook, CleanupHook)

        # All of the hook should satisfy the ResettableHook interface
        for hook in self.hooks:
            self.assertIsInstance(hook, ResettableHook)

    def _create_test_functions(self, event_name_prefix, *hooks):
        """Hook two test functions: one that does nothing, and other that raises a RuntimeError."""

        @self.auditor.audit(f"{event_name_prefix}.no_error", hooks=hooks)
        def no_error():
            return

        @self.auditor.audit(f"{event_name_prefix}.error", hooks=hooks)
        def error():
            raise RuntimeError()

        return no_error, error

    def test_hooks_a_and_b(self):
        # Tests for _HookA + _HookB
        nerr, err = self._create_test_functions("hook_ab", self.hook_a, self.hook_b)
        with self.auditor.start_auditing(reset_hooks=True):
            nerr()
            self.assertEqual(self.hook_a.counter, 1)
            self.assertEqual(self.hook_b.counter, 1)

            try:
                err()
                self.fail("err() did not raise RuntimeError")
            except RuntimeError as ex:
                self.ex = ex

            self.assertEqual(self.hook_a.counter, 1)
            self.assertEqual(self.hook_b.counter, 2)
            self.assertEqual(self.hook_b.exception[0], RuntimeError)

    def test_hooks_b_and_c(self):
        # Tests for _HookC + _HookB
        nerr, err = self._create_test_functions("hook_cb", self.hook_c, self.hook_b)
        with self.auditor.start_auditing(reset_hooks=True):
            with self.assertRaises(AssertionError):
                nerr()

            # Since the error was raised in the posthooks, and not in the prehooks or in
            # the wrapped function, 'exception' should be None when hook_b.cleanup and
            # hook_c.cleanup are called.
            self.assertEqual(self.hook_b.counter, 1)
            self.assertEqual(self.hook_c.counter, 1)
            self.assertEqual(self.hook_b.exception, (None, None, None))
            self.assertEqual(self.hook_c.exception, (None, None, None))

            # Despite the fact that an error is raised in the posthook, we should prioritize
            # the error that was raised by the wrapped function.
            try:
                err()
                self.fail("err() did not raise RuntimeError")
            except RuntimeError as ex:
                self.ex = ex
            self.assertEqual(self.hook_b.counter, 2)
            self.assertEqual(self.hook_c.counter, 2)
            self.assertEqual(self.hook_b.exception[0], RuntimeError)
            self.assertEqual(self.hook_c.exception[0], RuntimeError)

    def test_hooks_b_and_d(self):
        # Tests for _HookD + _HookB, and _HookB + _HookD
        #
        # For these test cases, we have to consider the fact that cleanup is only called on
        # the hooks whose prehooks were executed. As a result:
        # - When we call _HookD before _HookB, we error out *before* reaching _HookB's prehook, so
        #   we never call cleanup on _HookB.
        # - When we call _HookD after _HookB, we error out *after* reaching _HookB's prehook, so
        #   cleanup *should* be called on _HookB.

        nerr, err = self._create_test_functions("hook_db", self.hook_d, self.hook_b)
        with self.auditor.start_auditing(reset_hooks=True):
            with self.assertRaises(AssertionError):
                nerr()
            self.assertEqual(self.hook_b.counter, 0)
            self.assertEqual(self.hook_d.counter, 0)
            self.assertEqual(self.hook_b.exception, None)
            self.assertEqual(self.hook_d.exception, None)

            with self.assertRaises(AssertionError):
                err()
            self.assertEqual(self.hook_b.counter, 0)
            self.assertEqual(self.hook_d.counter, 0)
            self.assertEqual(self.hook_b.exception, None)
            self.assertEqual(self.hook_d.exception, None)

        nerr, err = self._create_test_functions("hook_bd", self.hook_b, self.hook_d)
        with self.auditor.start_auditing(reset_hooks=True):
            try:
                nerr()
                self.fail("nerr() did not raise AssertionError")
            except AssertionError as ex:
                self.ex = ex
            self.assertEqual(self.hook_b.counter, 1)
            self.assertEqual(self.hook_d.counter, 0)
            self.assertEqual(self.hook_b.exception[0], AssertionError)
            self.assertEqual(self.hook_d.exception, None)

            try:
                err()
                self.fail("err() did not raise AssertionError")
            except AssertionError as ex:
                self.ex = ex
            self.assertEqual(self.hook_b.counter, 2)
            self.assertEqual(self.hook_d.counter, 0)
            self.assertEqual(self.hook_b.exception[0], AssertionError)
            self.assertEqual(self.hook_d.exception, None)



import sys
import seagrass._typing as t
from contextvars import ContextVar, Token
from seagrass.base import CleanupHook
from types import FrameType

# Global variables that keep track of whether or not a TracingHook already exists.
_tracing_hook_exists: ContextVar[bool] = ContextVar(
    "_tracing_hook_exists", default=False
)
_tracing_hook: ContextVar[t.Optional["TracingHook.TraceFunc"]] = ContextVar(
    "_tracing_hook", default=None
)

TracingHookContext = t.Tuple[t.Optional[str], Token, Token]


class TracingHook(CleanupHook[TracingHookContext]):
    """Seagrass hook wrapper for tracing functions.

    **Example:** the code snippet below defines a new hook from
    :py:class:`~seagrass.hooks.TracingHook` that checks each frame to see if ``MY_VAR`` is defined
    locally, and if it is, it records ``MY_VAR``'s value.

    .. testsetup:: tracing-hook-example

        from seagrass._docs import configure_logging
        configure_logging()

    .. doctest:: tracing-hook-example

        >>> import seagrass

        >>> from seagrass.hooks import TracingHook

        >>> def tracefunc(frame, event, arg):
        ...     if "MY_VAR" in frame.f_locals:
        ...         MY_VAR = frame.f_locals["MY_VAR"]
        ...         logger = seagrass.get_audit_logger(None)
        ...         if logger is not None:
        ...             logger.info(f"Found MY_VAR={MY_VAR!r}")
        ...     return tracefunc

        >>> hook = TracingHook(tracefunc)

        >>> @seagrass.audit(seagrass.auto, hooks=[hook])
        ... def example():
        ...     MY_VAR = 100
        ...     MY_VAR = "hello, world!"

        >>> with seagrass.start_auditing():
        ...     example()
        {"message": "Found MY_VAR=100", "seagrass": {"event": "example"}, "level": "INFO"}
        {"message": "Found MY_VAR='hello, world!'", "seagrass": {"event": "example"}, "level": "INFO"}
    """

    class TraceFunc(t.Protocol):
        """A tracing function set by ``sys.settrace`` or ``threading.settrace``. The arguments to this
        function are the same as those to the function accepted by ``sys.settrace``."""

        def __call__(
            self, frame: FrameType, event: str, arg: t.Any
        ) -> t.Optional["TracingHook.TraceFunc"]:
            ...  # pragma: no cover

    def __init__(self, tracefunc: TraceFunc) -> None:
        """Create a new TracingHook.

        :param TraceFunc tracefunc: the function that should be used to perform tracing via
            `sys.settrace`_.

        .. _sys.settrace: https://docs.python.org/3/library/sys.html#sys.settrace
        """
        self.tracefunc = tracefunc

    # High prehook/posthook priority since we generally don't want to trace other
    # Seagrass hooks
    priority: int = 15

    __current_event: t.Optional[str] = None
    __is_active: bool = False

    @property
    def is_active(self) -> bool:
        """Return whether or not the hook is currently active (i.e., whether a Seagrass event that
        uses the hook is currently executing.)"""
        return self.__is_active

    @property
    def current_event(self) -> t.Optional[str]:
        """Return the current Seagrass event that is being executed."""
        return self.__current_event

    @staticmethod
    def get_current_tracing_hook():
        """Get the current global tracing hook."""
        return _tracing_hook.get()

    def __create_tracefunc(
        self,
        func: t.Optional["TraceFunc"],
    ) -> "TraceFunc":
        """A wrapper around the tracefunc function. This is the function that actually gets added
        with sys.settrace."""

        def tracefunc(
            frame: FrameType, event: str, arg: t.Any
        ) -> "TracingHook.TraceFunc":
            if func is not None and self.is_active:
                return self.__create_tracefunc(func(frame, event, arg))
            else:
                return self.__create_tracefunc(None)

        return tracefunc

    def prehook(
        self, event_name: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> TracingHookContext:

        # Check whether another TracingHook is already active
        exists_token = _tracing_hook_exists.set(True)
        if exists_token.old_value != Token.MISSING and not self.is_active:
            _tracing_hook_exists.reset(exists_token)
            current_hook = _tracing_hook.get()
            raise ValueError(
                f"Only one TracingHook can be active at a time (current tracing hook = {current_hook!r})"
            )

        tracefunc_token = _tracing_hook.set(self.__create_tracefunc(self.tracefunc))
        old_event = self.__current_event
        self.__current_event = event_name
        self.__is_active = True

        sys.settrace(_tracing_hook.get())

        return old_event, exists_token, tracefunc_token

    def cleanup(
        self,
        event_name: str,
        context: TracingHookContext,
        exc: t.Tuple[t.Any, ...],
    ) -> None:
        old_event, exists_token, tracefunc_token = context

        self.__current_event = old_event
        self.__is_active = self.current_event is not None
        _tracing_hook_exists.reset(exists_token)
        _tracing_hook.reset(tracefunc_token)

        sys.settrace(_tracing_hook.get())



import sys
import seagrass._typing as t
from contextlib import contextmanager
from functools import wraps
from seagrass.auditor import get_audit_logger
from seagrass.base import ProtoHook

# Type variable used to represent value returned from a function
R = t.TypeVar("R")


class RuntimeAuditHook(ProtoHook[t.Optional[str]]):
    """A hook that executes its body as a Python runtime audit hook, in accordance with `PEP 578`_.

    .. note::
        :py:class:`~seagrass.hooks.RuntimeAuditHook` is only supported for Python versions >= 3.8

    **Examples:** in the code below, ``RuntimeEventCounterHook`` is a class derived from the
    ``RuntimeAuditHook`` base class that prints every time a runtime audit event is triggered.

    .. testsetup:: runtime-audit-hook-example

        from seagrass.auditor import Auditor
        auditor = Auditor()

    .. doctest:: runtime-audit-hook-example
        :pyversion: >= 3.8

        >>> import sys

        >>> from seagrass.hooks import RuntimeAuditHook

        >>> class RuntimeEventCounterHook(RuntimeAuditHook):
        ...     def __init__(self):
        ...         super().__init__(self.sys_hook)
        ...
        ...     def sys_hook(self, event, args):
        ...         print(f"Encountered event={event!r} with args={args}")
        ...

        >>> hook = RuntimeEventCounterHook()

        >>> @auditor.audit("my_event", hooks=[hook])
        ... def my_event(*args):
        ...     sys.audit("sys.my_event", *args)

        >>> with auditor.start_auditing():
        ...     my_event(42, "hello, world")
        Encountered event='sys.my_event' with args=(42, 'hello, world')

    .. _PEP 578: https://www.python.org/dev/peps/pep-0578/
    """

    # Default value of the propagate_errors attribute
    PROPAGATE_ERRORS_DEFAULT: t.Final[bool] = True

    # Whether to propagate errors from sys_hook.
    propagate_errors: bool = PROPAGATE_ERRORS_DEFAULT

    __current_event: t.Optional[str] = None
    __is_active: bool

    @property
    def is_active(self) -> bool:
        """Return whether or not the hook is currently active (i.e., whether a Seagrass event that
        uses the hook is currently executing.)"""
        return self.__is_active

    @property
    def current_event(self) -> t.Optional[str]:
        """Returns the current Seagrass event being executed that's hooked by this function. If no
        events using this hook are being executed, ``current_event`` is ``None``."""
        return self.__current_event

    def __update_properties(self) -> None:
        self.__is_active = self.current_event is not None

    def __update(func: t.Callable[..., R]) -> t.Callable[..., R]:  # type: ignore[misc]
        # NOTE: mypy will flag this as erroneous because it is a non-static method that doesn't
        # include the argument 'self'
        # Ref: https://github.com/python/mypy/issues/7778
        """Function decorator that causes functions to reset the current_event and is_active
        properties every time it gets called."""

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            finally:
                self.__update_properties()

        return wrapper

    def __init__(
        self,
        sys_hook: t.Callable[[str, t.Tuple[t.Any, ...]], None],
        propagate_errors: t.Optional[bool] = None,
        traceable: bool = False,
    ) -> None:
        if not hasattr(sys, "audit"):
            raise NotImplementedError(
                "RuntimeAuditHook is not supported for Python versions that don't "
                "include sys.audit and sys.addaudithook"
            )

        self.sys_hook = sys_hook

        if propagate_errors is not None:
            self.propagate_errors = propagate_errors
        self.__update_properties()

        hook = self.__create_sys_hook()

        # Per https://docs.python.org/3/library/sys.html#sys.addaudithook, audit hooks are not traced
        # unless __cantrace__ is set to True.
        if traceable:
            hook.__cantrace__ = True  # type: ignore[attr-defined]

        # Add the runtime audit hook after initializing the properties since the hook will in most
        # cases use some of the properties of the RuntimeAuditHook.
        sys.addaudithook(hook)

    def __create_sys_hook(self) -> t.Callable[[str, t.Tuple[t.Any, ...]], None]:
        """Creates wrapper around the sys_hook abstract method that first checks whether the hook
        is currently active before it executes anything. This is the function that actually
        gets added with sys.addaudithook, not sys_hook."""

        def __sys_hook(event: str, args: t.Tuple[t.Any, ...]) -> None:
            if self.is_active:
                try:
                    self.sys_hook(event, args)
                except Exception as ex:
                    if self.propagate_errors:
                        raise ex
                    else:
                        logger = get_audit_logger(None)
                        if logger is not None:
                            # Temporarily disable the hook, since emitting a log could create new
                            # runtime events. In some cases this could lead to an infinite recursion.
                            with self.__disable_runtime_hook():
                                logger.error(
                                    "%s raised in %s.sys_hook: %s",
                                    ex.__class__.__name__,
                                    self.__class__.__name__,
                                    ex,
                                )

        return __sys_hook

    @contextmanager
    def __disable_runtime_hook(self) -> t.Iterator[None]:
        """Temporarily the runtime hook."""
        is_active = self.__is_active
        self.__is_active = False
        try:
            yield None
        finally:
            self.__is_active = is_active

    @__update
    def prehook(
        self, event: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> t.Optional[str]:
        old_event = self.__current_event
        self.__current_event = event
        return old_event

    def posthook(self, event: str, result: t.Any, context: t.Optional[str]) -> None:
        pass

    @__update
    def cleanup(
        self, event: str, context: t.Optional[str], exc: t.Optional[Exception]
    ) -> None:
        self.__current_event = context


RuntimeAuditHook.__init__.__doc__ = f"""\
Initialize the RuntimeAuditHook.

:param Optional[bool] propagate_errors: if not equal to ``None``, set the ``propagate_errors``
    attribute of the hook, which controls whether errors raised by :py:meth:`sys_hook` are
    propagated. Hooks for which ``propagate_errors`` is ``False`` will log errors raised by
    :py:meth:`sys_hook`, but won't raise them.

    By default, ``hook.propagate_errors = {RuntimeAuditHook.PROPAGATE_ERRORS_DEFAULT}`` for
    RuntimeAuditHooks.
"""



import sys
import seagrass._typing as t
from abc import abstractmethod, ABCMeta
from contextlib import contextmanager, ExitStack
from functools import wraps
from seagrass.base import ProtoHook

from .contexts import EventData, HookExecutionContext, current_event

# Type variable used to represent the value returned by an event function
R = t.TypeVar("R")

# A type variable used to represent the function wrapped by an Event.
F = t.Callable[..., t.Any]

# Function decorators for the __call__ method of events


def _check_enabled(func: t.Callable) -> t.Callable:
    """Function decorator that checks whether the event is enabled before entering its body.
    If the event is not enabled, it will skip over executing the body of the function and
    just execute the wrapped function."""

    @wraps(func)
    def wrapper(self: Event, *args, **kwargs):
        if not self.enabled:
            return self.func(*args, **kwargs)
        else:
            return func(self, *args, **kwargs)

    return wrapper


class Event(metaclass=ABCMeta):
    """Defines an event that is under audit. The event wraps around a function; instead of calling
    the function, we call the event, which first triggers any prehooks, *then* calls the function,
    and then triggers posthooks."""

    # Use __slots__ since feasibly users may want to create a large
    # number of events
    __slots__ = [
        "func",
        "enabled",
        "name",
        "raise_runtime_events",
        "hooks",
        "prehook_audit_event_name",
        "posthook_audit_event_name",
        "_hook_execution_order",
    ]

    enabled: bool
    name: str
    raise_runtime_events: bool
    hooks: t.List[ProtoHook]
    prehook_audit_event_name: str
    posthook_audit_event_name: str
    _hook_execution_order: t.List[int]

    def __init__(
        self,
        func: F,
        name: str,
        enabled: bool = True,
        hooks: t.List[ProtoHook] = [],
        raise_runtime_events: bool = False,
        prehook_audit_event_name: t.Optional[str] = None,
        posthook_audit_event_name: t.Optional[str] = None,
    ) -> None:
        """Create a new Event.

        :param Callable[[...],Any] func: the function being wrapped by this event.
        :param str name: the name of the event.
        :param bool enabled: whether to enable the event.
        :param List[ProtoHook] hooks: a list of all of the hooks that should be called whenever
            the event is triggered.
        :param bool raise_runtime_events: if ``True``, two `Python runtime audit events`_ are raised
            using `sys.audit`_ before and after running the function wrapped by the event.

            .. note::
                This parameter is only supported for Python version >= 3.8.

        :param Optional[str] prehook_audit_event_name: the name of the runtime audit event
            that should be raised *before* calling the wrapped function. If set to ``None``,
            the audit event is automatically named ``f"prehook:{name}"``. This parameter is
            ignored if ``raise_runtime_events`` is ``False``.
        :param Optional[str] posthook_audit_event_name: the name of the runtime audit event
            that should be raised *after* calling the wrapped function. If set to ``None``,
            the audit event is automatically named ``f"posthook:{name}"``. This parameter is
            ignored if ``raise_runtime_events`` is ``False``.

        .. _Python runtime audit events: https://www.python.org/dev/peps/pep-0578/
        .. _sys.audit: https://docs.python.org/3/library/sys.html#sys.audit
        """
        self.func: F = func
        self.enabled = enabled
        self.name = name
        self.hooks = []

        if raise_runtime_events:
            # Check that the Python version supports audit hooks
            if not hasattr(sys, "audit"):
                raise NotImplementedError(
                    "Runtime audit events are not supported for Python versions that don't "
                    "include sys.audit and sys.addaudithook"
                )

        self.raise_runtime_events = raise_runtime_events

        if prehook_audit_event_name is None:
            prehook_audit_event_name = f"prehook:{name}"
        if posthook_audit_event_name is None:
            posthook_audit_event_name = f"posthook:{name}"

        self.prehook_audit_event_name = prehook_audit_event_name
        self.posthook_audit_event_name = posthook_audit_event_name

        self.add_hooks(*hooks)

    def add_hooks(self, *hooks: ProtoHook) -> None:
        """Add new hooks to the event.

        :param ProtoHook hooks: the hooks to add to the event.
        """
        for hook in hooks:
            self.hooks.append(hook)

        # Since we've updated the list of hooks, we need to re-determine the order
        # in which the hooks should be executed.
        self._set_hook_execution_order()

    def _set_hook_execution_order(self) -> None:
        """Determine the order in which the events' hooks should be executed."""
        # - Prehooks are ordered by ascending priority, then ascending list position
        # - Posthooks are ordered by descending priority, then descending list position
        self._hook_execution_order = sorted(
            range(len(self.hooks)), key=lambda i: (self.hooks[i].priority, i)
        )

    @contextmanager
    def _event_context(
        self, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> t.Iterator[EventData]:
        token = current_event.set(self.name)
        event_data = EventData(self.name, args, kwargs)

        try:
            if self.raise_runtime_events:
                sys.audit(self.prehook_audit_event_name, args, kwargs)

            with ExitStack() as stack:
                for hook_num in self._hook_execution_order:
                    hook = self.hooks[hook_num]
                    if hook.enabled:
                        stack.enter_context(HookExecutionContext(hook, event_data))

                yield event_data

        finally:
            try:
                if self.raise_runtime_events:
                    sys.audit(self.posthook_audit_event_name, event_data.result)
            finally:
                current_event.reset(token)

    @abstractmethod
    def __call__(self, *args, **kwargs) -> t.Any:
        ...  # pragma: no cover


class SyncEvent(Event):
    @_check_enabled
    def __call__(self, *args, **kwargs) -> t.Any:
        """Call the function wrapped by the Event. If the event is enabled, its prehooks and
        posthooks are executed before and after the execution of the wrapped function.

        :param args: the arguments to pass to the wrapped function.
        :param kwargs: the keyword arguments to pass to the wrapped function.
        """

        with self._event_context(args, kwargs) as event_data:
            event_data.result = self.func(*args, **kwargs)

        if isinstance(event_data.result, t.Missing):
            # This point should never be reached
            raise ValueError("Event result was not recorded")  # pragma: no cover

        return event_data.result


class AsyncEvent(Event):
    @_check_enabled
    async def __call__(self, *args, **kwargs) -> t.Any:
        with self._event_context(args, kwargs) as event_data:
            event_data.result = await self.func(*args, **kwargs)

        if isinstance(event_data.result, t.Missing):
            # This point should never be reached
            raise ValueError("Event result was not recorded")  # pragma: no cover

        return event_data.result


AsyncEvent.__call__.__doc__ = SyncEvent.__call__.__doc__



# Tests for Auditor creation and basic functionality

import logging
import seagrass
import unittest
from io import StringIO
from seagrass import Auditor, get_audit_logger
from seagrass.base import ProtoHook
from seagrass.errors import EventNotFoundError
from test.utils import SeagrassTestCaseMixin


class CreateAuditorTestCase(unittest.TestCase):
    """Tests for creating a new Auditor instance."""

    def _clear_logging_output(self):
        # Helper function to clear output buffer used for testing logging
        self.logging_output.seek(0)
        self.logging_output.truncate()

    def _configure_logger(self, logger: logging.Logger):
        # Set the testing configuration for an input logger
        logger.setLevel(logging.INFO)

        fh = logging.StreamHandler(self.logging_output)
        fh.setLevel(logging.DEBUG)

        formatter = logging.Formatter("(%(levelname)s) %(name)s: %(message)s")
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    def setUp(self):
        self.logging_output = StringIO()

    def test_use_custom_logger_for_auditor(self):
        # If a logger is explicitly specified, Auditor should use that logger instead
        # of the default. The logger can either be provided as a logging.Logger
        # instance or as the name of a logger.
        self._configure_logger(logging.getLogger("test_logger"))

        auditor = Auditor(logger="test_logger")
        auditor.logger.info("Hello, world!")
        auditor.logger.debug("This message shouldn't appear")

        output = self.logging_output.getvalue()
        self.assertEqual(output, "(INFO) test_logger: Hello, world!\n")

        self._clear_logging_output()


class SimpleAuditorFunctionsTestCase(SeagrassTestCaseMixin, unittest.TestCase):
    def test_define_new_event(self):
        # Define a new event and ensure that it gets added to the auditor's events
        # dictionary and event_wrapper dictionary
        @self.auditor.audit("test.foo")
        def foo():
            return

        self.assertIn("test.foo", self.auditor.events)
        self.assertIn("test.foo", self.auditor.event_wrappers)

    def test_define_two_events_with_the_same_name(self):
        @self.auditor.audit("test.foo")
        def foo_1():
            return

        with self.assertRaises(ValueError):

            @self.auditor.audit("test.foo")
            def foo_2():
                return

    def test_create_empty_event(self):
        # Create a new audit event that doesn't wrap any existing function.
        with self.assertRaises(EventNotFoundError):
            self.auditor.raise_event("test.signal", 1, 2, name="Alice")

        class TestHook(ProtoHook):
            def __init__(self):
                self.last_prehook_args = self.last_posthook_args = None

            def prehook(self, event_name, args, kwargs):
                self.last_prehook_args = (event_name, args, kwargs)

            def posthook(self, event_name, result, context):
                self.last_posthook_args = (event_name, result)

        hook = TestHook()
        self.auditor.create_event("test.signal", hooks=[hook])

        # Event shouldn't be triggered outside of an auditing context
        self.auditor.raise_event("test.signal", 1, 2, name="Alice")
        self.assertEqual(hook.last_prehook_args, None)
        self.assertEqual(hook.last_posthook_args, None)

        with self.auditor.start_auditing():
            self.auditor.raise_event("test.signal", 1, 2, name="Alice")

        self.assertEqual(
            hook.last_prehook_args, ("test.signal", (1, 2), {"name": "Alice"})
        )
        self.assertEqual(hook.last_posthook_args, ("test.signal", None))

    def test_raise_event_cumsum(self):
        # Insert an audit event into the function my_sum so that we can monitor the internal
        # state of the function as it's executing. In this case, we'll be retrieving the
        # cumulative sum at various points in time.
        def my_sum(*args):
            total = 0.0
            for arg in args:
                self.auditor.raise_event("my_sum.cumsum", total)
                total += arg

        class MySumHook(ProtoHook[None]):
            def __init__(self):
                self.cumsums = []

            def prehook(self, event_name, args, kwargs):
                self.cumsums.append(args[0])

            def posthook(self, *args):
                pass

        hook = MySumHook()
        self.auditor.create_event("my_sum.cumsum", hooks=[hook])

        with self.auditor.start_auditing():
            my_sum(1, 2, 3, 4)

        self.assertEqual(hook.cumsums, [0.0, 1.0, 3.0, 6.0])

    def test_get_audit_logger(self):
        # Tests for the get_audit_logger function

        outer_auditor = Auditor(logger="outer")
        inner_auditor = Auditor(logger="inner")
        self.assertNotEqual(outer_auditor.logger, inner_auditor.logger)

        # Outside of an auditing context, get_audit_logger should return the default value
        # (or otherwise raise a LookupError)
        self.assertEqual(get_audit_logger(None), None)
        with self.assertRaises(LookupError):
            get_audit_logger()

        # Within an auditing context, get_audit_logger() should return the logger
        # for the most recent auditing context.
        with outer_auditor.start_auditing():
            self.assertEqual(get_audit_logger(), outer_auditor.logger)

            with inner_auditor.start_auditing():
                self.assertEqual(get_audit_logger(), inner_auditor.logger)

            self.assertEqual(get_audit_logger(), outer_auditor.logger)

        # Now that we're back outside of an auditing context, get_audit_logger()
        # should once again raise an error/return the default
        self.assertEqual(get_audit_logger(None), None)
        with self.assertRaises(LookupError):
            get_audit_logger()

    def test_filter_events(self):
        hook = seagrass.hooks.CounterHook()
        self.auditor.create_event("test.foo", hooks=[hook])
        self.auditor.create_event("bar", hooks=[hook])

        ec = hook.event_counter

        def raise_events():
            for event in ("test.foo", "bar"):
                self.auditor.raise_event(event)

        with self.auditor.start_auditing(
            filter=lambda event: event.startswith("test"), reset_hooks=True
        ):
            raise_events()
            self.assertEqual(ec["test.foo"], 1)
            self.assertEqual(ec["bar"], 0)

        # After we leave the auditing context, the filter should be reset.
        with self.auditor.start_auditing(reset_hooks=True):
            raise_events()
            self.assertEqual(ec["test.foo"], 1)
            self.assertEqual(ec["bar"], 1)

        # It should also be possible to modify the event filter using the setter. This setter is not
        # context-dependent.
        self.auditor.event_filter = lambda event: event.startswith("test")

        with self.auditor.start_auditing(reset_hooks=True):
            raise_events()
            self.assertEqual(ec["test.foo"], 1)
            self.assertEqual(ec["bar"], 0)

        with self.auditor.start_auditing(reset_hooks=True):
            raise_events()
            self.assertEqual(ec["test.foo"], 1)
            self.assertEqual(ec["bar"], 0)

        # However, once we call reset_filter, all events should start executing as normal again
        self.auditor.reset_filter()
        with self.auditor.start_auditing(reset_hooks=True):
            raise_events()
            self.assertEqual(ec["test.foo"], 1)
            self.assertEqual(ec["bar"], 1)


class GlobalAuditorTestCase(unittest.TestCase):
    """Tests for using the global auditor instance."""

    def test_use_counter_hook_with_global_auditor(self):
        hook = seagrass.hooks.CounterHook()

        @seagrass.audit("test.foo", hooks=[hook])
        def foo():
            pass

        with seagrass.start_auditing():
            for _ in range(10):
                foo()

        self.assertEqual(hook.event_counter["test.foo"], 10)

    def test_create_global_auditor_in_context(self):
        hook = seagrass.hooks.CounterHook()

        with seagrass.create_global_auditor() as auditor:
            self.assertEqual(seagrass.global_auditor(), auditor)

            # Any events created inside of this context should be created with the new auditor
            @seagrass.audit("test.foo", hooks=[hook])
            def foo():
                pass

        # Outside of the context, the auditor should be set back to the original
        # global auditor.
        self.assertNotEqual(seagrass.global_auditor(), auditor)

        # Events created inside of the context should only be raised with the newly-created
        # auditor.
        with seagrass.start_auditing(reset_hooks=True):
            foo()
            self.assertEqual(hook.event_counter["test.foo"], 0)

        with auditor.start_auditing(reset_hooks=True):
            foo()
            self.assertEqual(hook.event_counter["test.foo"], 1)

        # It should also be possible to create an Auditor and pass it into create_global_auditor
        auditor = Auditor()
        with seagrass.create_global_auditor(auditor):
            self.assertEqual(seagrass.global_auditor(), auditor)

        self.assertNotEqual(seagrass.global_auditor(), auditor)


if __name__ == "__main__":
    unittest.main()



# flake8: noqa: F401
import functools as _functools
import seagrass._typing as t
from ._logging import SeagrassLogFilter
from .auditor import Auditor, get_audit_logger, DEFAULT_LOGGER_NAME
from .events import get_current_event
from . import base, errors, events, hooks
from contextvars import ContextVar as _ContextVar
from logging import Logger as _Logger

# "Global auditor" that can be used to audit events without having to create an
# auditor first.
_GLOBAL_AUDITOR: _ContextVar[Auditor] = _ContextVar(
    "_GLOBAL_SEAGRASS_AUDITOR", default=Auditor()
)


def global_auditor() -> Auditor:
    """Return the global Seagrass auditor."""
    return _GLOBAL_AUDITOR.get()


def auto(func: t.Callable) -> str:
    """Automatically generate an event name for a function being audited.

    **Examples:**

        .. testsetup:: auto-doctests

            from seagrass import Auditor
            from seagrass._docs import configure_logging
            configure_logging()
            auditor = Auditor()

        .. doctest:: auto-doctests

            >>> from seagrass import auto

            >>> from time import sleep

            >>> event = auditor.audit(auto, sleep)

            >>> event.__event_name__
            'time.sleep'

            >>> from pathlib import Path

            >>> event = auditor.audit(auto, Path.home)

            >>> event.__event_name__
            'pathlib.Path.home'
    """
    if func.__module__ is None:
        return func.__qualname__    # type: ignore[unreachable]

    return f"{func.__module__}.{func.__qualname__}"


_F = t.TypeVar("_F", bound=t.Callable)

# Export parts of the external API of the global Auditor instance from the module
_EXPORTED_AUDITOR_ATTRIBUTES: t.List[str] = []


def _auditor_func(get_func: t.Callable[[], _F]) -> _F:
    """Wrap a function that retrieves a method of the global auditor so that it can be called as
    though it were the original method."""
    # Get the current global auditor's current instance of the input function
    # so that we can copy its name, annotations, documentation, etc.
    func = get_func()

    # Add the function's name to the _EXPORTED_AUDITOR_ATTRIBUTES list
    _EXPORTED_AUDITOR_ATTRIBUTES.append(func.__name__)

    @_functools.wraps(func)
    def wrapper(*args, **kwargs):
        return get_func()(*args, **kwargs)

    return t.cast(_F, wrapper)


audit = _auditor_func(lambda: global_auditor().audit)
async_audit = _auditor_func(lambda: global_auditor().audit)
create_event = _auditor_func(lambda: global_auditor().create_event)
raise_event = _auditor_func(lambda: global_auditor().raise_event)
toggle_event = _auditor_func(lambda: global_auditor().toggle_event)
toggle_auditing = _auditor_func(lambda: global_auditor().toggle_auditing)
start_auditing = _auditor_func(lambda: global_auditor().start_auditing)
add_hooks = _auditor_func(lambda: global_auditor().add_hooks)
reset_hooks = _auditor_func(lambda: global_auditor().reset_hooks)

# Remaining attributes are properties that need to be retrieved using the module __getattr__
# function
logger: _Logger

_EXPORTED_AUDITOR_ATTRIBUTES += ["logger"]


class create_global_auditor(t.ContextManager[Auditor]):
    """Create a context with a new global Auditor (as returned by the ``global_auditor()``
    function.) This is useful for when you want to import a module that uses Seagrass but
    don't want to add its events to the current global Auditor.

    If an Auditor is passed into this function, it will be used as the global auditor within the
    created context. Otherwise, a new Auditor instance will be created.

    :param Optional[Auditor] auditor: the :py:class:`seagrass.Auditor` instance that should be used
        as the global auditor. If no auditor is provided, a new one will be created.

    .. doctest:: create_global_auditor_doctests

        >>> import seagrass

        >>> from seagrass.hooks import LoggingHook

        >>> hook = LoggingHook(prehook_msg=lambda event, *args: f"called {event}")

        >>> with seagrass.create_global_auditor() as auditor:
        ...     @seagrass.audit("my_event", hooks=[hook])
        ...     def my_event():
        ...         pass

        >>> with seagrass.start_auditing():
        ...     my_event()

        >>> with auditor.start_auditing():
        ...     my_event()
        {"message": "called my_event", "seagrass": {"event": "my_event"}, "level": "DEBUG"}
    """

    def __init__(self, auditor: t.Optional[Auditor] = None) -> None:
        if auditor is None:
            self.new_auditor = Auditor()
        else:
            self.new_auditor = auditor

    def __enter__(self) -> Auditor:
        self.auditor_token = _GLOBAL_AUDITOR.set(self.new_auditor)
        return self.new_auditor

    def __exit__(self, *args) -> None:
        _GLOBAL_AUDITOR.reset(self.auditor_token)


__all__ = [
    "base",
    "errors",
    "events",
    "hooks",
    "DEFAULT_LOGGER_NAME",
    "Auditor",
    "SeagrassLogFilter",
    "get_audit_logger",
    "get_current_event",
    "global_auditor",
    "create_global_auditor",
    "auto",
]

__all__ += _EXPORTED_AUDITOR_ATTRIBUTES


def __getattr__(attr) -> t.Any:
    if attr == "logger":
        return global_auditor().logger
    else:
        raise AttributeError(f"module {__name__!r} has no attribute {attr!r}")



import seagrass._typing as _t
from contextvars import ContextVar, Token

current_hook_ctx: ContextVar["HookContext"] = ContextVar("seagrass_hook_ctx")

if _t.TYPE_CHECKING:
    import sys

    if sys.version_info >= (3, 9):
        CtxType = Token["HookContext"]
    else:
        CtxType = Token


class HookContext:
    """Logging context provided by a hook that gets attached to logs."""

    __slots__ = ["name", "args", "previous_ctx"]

    name: str
    args: _t.Dict[str, _t.Any]
    previous_ctx: _t.Optional["CtxType"]

    def __init__(self, name: str, args: _t.Dict[str, _t.Any]) -> None:
        """Create a new ``HookContext`` instance."""

        self.name = name
        self.args = args
        self.previous_ctx = None

    def __enter__(self) -> "HookContext":
        self.previous_ctx = current_hook_ctx.set(self)
        return self

    def __exit__(self, *args) -> None:
        if self.previous_ctx is not None:
            current_hook_ctx.reset(self.previous_ctx)
        self.previous_ctx = None


T = _t.TypeVar("T")


@_t.overload
def get_hook_context() -> HookContext:
    ...  # pragma: no cover


@_t.overload
def get_hook_context(default: T) -> _t.Union[HookContext, T]:
    ...  # pragma: no cover


def get_hook_context(default: _t.Maybe[T] = _t.MISSING) -> _t.Union[HookContext, T]:
    """Retrieve the logging context provided by an executing hook.

    :raises LookupError: if no hook is currently being executed, and ``default`` is not
        specified.
    """
    if isinstance(default, _t.Missing):
        return current_hook_ctx.get()
    else:
        return current_hook_ctx.get(default)



import seagrass._typing as t
from collections import defaultdict
from logging import Logger
from .runtime_audit_hook import RuntimeAuditHook


class FileOpenInfo(t.NamedTuple):
    filename: str
    mode: str
    flags: int


class FileOpenHook(RuntimeAuditHook):
    """An event hook for tracking calls to the Python standard
    library's `open` function."""

    # Give this hook slightly higher priority by default so that
    # we can avoid counting calls to open that occur in other
    # hooks.
    priority: int = 3

    file_open_counter: t.DefaultDict[str, t.Counter[FileOpenInfo]]

    def __init__(self, **kwargs) -> None:
        super().__init__(self.sys_hook, **kwargs)
        self.file_open_counter = defaultdict(t.Counter[FileOpenInfo])

    def sys_hook(self, runtime_event: str, args: t.Tuple[t.Any, ...]) -> None:
        if runtime_event == "open" and self.current_event is not None:
            filename, mode, flags = args
            info = FileOpenInfo(filename, mode, flags)
            self.file_open_counter[self.current_event][info] += 1

    def reset(self) -> None:
        self.file_open_counter.clear()

    def log_results(self, logger: Logger) -> None:
        logger.info("%s results (file opened, count):", self.__class__.__name__)
        for event in sorted(self.file_open_counter):
            logger.info("  event %s:", event)
            for (info, count) in self.file_open_counter[event].items():
                logger.info(
                    "    %s (mode=%s, flags=%s): opened %d times",
                    info.filename,
                    info.mode,
                    hex(info.flags),
                    count,
                )



import logging
import seagrass._typing as t
from types import TracebackType

# Type variable for contexts returned by prehooks
C = t.TypeVar("C")

DEFAULT_PRIORITY: int = 0


class ProtoHook(t.Protocol[C]):
    """Interface for hooks that can be used by Seagrass. New Seagrass hooks must define all of
    the properties and methods required for this class.

    Here's an example of a minimal hook that satisfies the ProtoHook interface. All this hook
    does is make an assertion that the first argument to a function wrapped by an audited
    event is a string.

    .. testsetup:: example_impl

        from seagrass import Auditor
        auditor = Auditor()

    .. doctest:: example_impl

        >>> from seagrass.base import ProtoHook

        >>> class TypeCheckHook(ProtoHook[None]):
        ...     def prehook(self, event_name, args, kwargs):
        ...         assert isinstance(args[0], str), "Input must be type str"
        ...
        ...     def posthook(self, event_name, result, context):
        ...         pass

        >>> @auditor.audit("event.say_hello", hooks=[TypeCheckHook()])
        ... def say_hello(name: str):
        ...     return f"Hello, {name}!"

        >>> with auditor.start_auditing():
        ...     say_hello("Alice")
        ...     say_hello(0)    # Should raise AssertionError   # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        AssertionError: Input must be type str
    """

    enabled: bool = True
    priority: int = DEFAULT_PRIORITY

    def prehook(
        self, event_name: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> C:
        """Run the prehook. The prehook is run at the beginning of the execution of
        an audited event, before the function wrapped by the event is run.

        :param str event_name: The name of the event that was triggered.
        :param Tuple[Any,...] args: A tuple of the arguments passed to the function wrapped by
            the event.
        :param Dict[str,Any] kwargs: A dictionary of the keyword arguments passed to the function
            wrapped by the event.
        :return: "context" data that can be used by the posthook.
        :rtype: C
        """

    def posthook(self, event_name: str, result: t.Any, context: C) -> None:
        """Run the posthook. The posthook is run at the end of the execution of
        an audited event, after the function wrapped by the event is run.

        :param str event_name: The name of the event that was triggered.
        :param Any result: The value that was returned by the event's wrapped function.
        :param C context: The context that was returned by the original call to ``prehook``.
        """

        # By default, the posthook function does nothing
        return


@t.runtime_checkable
class LogResultsHook(ProtoHook, t.Protocol):
    """A protocol class for hooks that support an additional `log_results` method that
    outputs the results of the hook."""

    def log_results(
        self,
        logger: logging.Logger,
    ) -> None:
        """Log results that have been accumulated by the hook using the provided logger.

        :param logging.Logger logger: the logger that should be used to output results.
        """


@t.runtime_checkable
class ResettableHook(ProtoHook, t.Protocol):
    """A protocol class for hooks that can be reset.

    **Examples:** here is a minimal example of a Seagrass hook that satisfies the
    ``ResettableHook`` interface. Every time the event ``"my_event"`` is raised,
    the hook prints the number of times the event has been raised so far and
    increments its counter.

    .. doctest::

       >>> from seagrass.base import ProtoHook, ResettableHook

       >>> class PrintEventHook(ProtoHook[None]):
       ...     def __init__(self):
       ...         self.reset()
       ...
       ...     def prehook(self, event_name, *args):
       ...         if event_name == "my_event":
       ...             self.event_counter += 1
       ...             print(f"my_event has be raised {self.event_counter} times")
       ...
       ...     def reset(self):
       ...         self.event_counter = 0
       ...

       >>> hook = PrintEventHook()

       >>> isinstance(hook, ResettableHook)
       True
    """

    def reset(self) -> None:
        """Reset the internal state of the hook."""


@t.runtime_checkable
class CleanupHook(ProtoHook[C], t.Protocol[C]):
    """A protocol class for hooks that have a 'cleanup' stage.

    Hooks that implement this interface are unconditionally 'cleaned up' after an event if their
    prehook was run during that event. Hooks that modify their internal state or create other
    side effects that need to be reset after the event is finished executing should implement
    this interface, in case an exception is thrown during the course of the event.
    """

    def cleanup(
        self,
        event_name: str,
        context: C,
        exc: t.Tuple[
            t.Optional[t.Type[BaseException]],
            t.Optional[BaseException],
            t.Optional[TracebackType],
        ],
    ) -> None:
        """Perform the hook's cleanup stage. The ``event_name`` and ``context`` are the same as
        those used by the ``posthook`` function. If an exception was thrown while executing the
        event it will be provided in the ``exception`` argument, otherwise, ``exception`` will
        be ``None``.

        If ``cleanup`` returns a boolean value, that value is used to decide whether to suppress
        any exceptions that were raised during event execution.

        :rtype: Optional[bool]
        """



# Additional tests for Sseagrass logging functionality

import json
import unittest
from seagrass import get_audit_logger
from test.utils import SeagrassTestCaseMixin


class AuditorLoggingTestCase(SeagrassTestCaseMixin, unittest.TestCase):
    def setUp(self):
        super().setUp()

    def test_log_events(self):
        @self.auditor.audit("myevent.foo")
        def foo():
            logger = get_audit_logger()
            if logger is not None:
                logger.info("hello, world!")

        @self.auditor.audit("myevent.bar")
        def bar():
            foo()
            logger = get_audit_logger()
            if logger is not None:
                logger.info("calling event bar")

        with self.auditor.start_auditing():
            bar()

        # Check that the event was written to the log
        output = self.logging_output.getvalue()[:-1].split("\n")
        output = [json.loads(o) for o in output]
        self.assertEqual(len(output), 2)
        self.assertTrue(all(o["level"] == "INFO" for o in output))

        self.assertEqual(output[0]["message"], "hello, world!")
        self.assertEqual(output[1]["message"], "calling event bar")

        self.assertEqual(output[0]["seagrass"]["event"], "myevent.foo")
        self.assertEqual(output[1]["seagrass"]["event"], "myevent.bar")

    def test_log_outside_of_event(self):
        self.auditor.logger.debug("log outside event")
        output = self.logging_output.getvalue()[:-1]
        output = json.loads(output)
        self.assertEqual(output["level"], "DEBUG")
        self.assertEqual(output["seagrass"]["event"], None)
        self.assertEqual(output["message"], "log outside event")


if __name__ == "__main__":
    unittest.main()



# Tests for the RuntimeAuditHook abstract base class.

import json
import sys
import tempfile
import unittest
from seagrass.base import CleanupHook
from seagrass.hooks import RuntimeAuditHook
from test.utils import HookTestCaseMixin, req_python_version


class RuntimeHookTestCaseMixin(HookTestCaseMixin):
    check_interfaces = (CleanupHook,)


# Example hooks that subclass RuntimeAuditHook


class FileOpenTestHook(RuntimeAuditHook):
    """A simple RuntimeAuditHook that tracks file opens. This is a simplified version of
    FileOpenHook."""

    def __init__(self):
        super().__init__(self.sys_hook, traceable=True)
        self.total_file_opens = 0

    def sys_hook(self, event_name, args):
        if event_name == "open":
            self.total_file_opens += 1
            self.opened_filename = args[0]
            self.opened_mode = args[1]


class ErroneousHook(RuntimeAuditHook):
    """A runtime audit hook that raises an error whenever sys_hook gets called."""

    def __init__(self):
        super().__init__(self.sys_hook, propagate_errors=False, traceable=True)

    def sys_hook(self, event, args):
        raise ValueError("my_test_message")


# Test cases


class FileOpenRuntimeHookTestCase(RuntimeHookTestCaseMixin, unittest.TestCase):
    """Build an auditing hook for tracking file opens out of RuntimeAuditHook, similar to
    FileOpenHook."""

    hook_gen = FileOpenTestHook

    # These test cases require the use of sys.audit and sys.addaudithook, so they're disabled
    # for Python versions < 3.8
    @req_python_version(min=(3, 8))
    def setUp(self):
        super().setUp()

    def test_hook_function(self):
        @self.auditor.audit("test.say_hello", hooks=[self.hook])
        def say_hello(filename) -> str:
            with open(filename, "w") as f:
                f.write("Hello!\n")

            with open(filename, "r") as f:
                return f.read()

        with tempfile.NamedTemporaryFile() as f:
            # Even though we're using sys.audit hooks, calls to say_hello should not
            # trigger the audit hook unless we're in an auditing context.
            say_hello(f.name)
            with self.auditor.start_auditing():
                say_hello(f.name)
            say_hello(f.name)

            self.assertEqual(self.hook.total_file_opens, 2)
            self.assertEqual(self.hook.opened_filename, f.name)
            self.assertEqual(self.hook.opened_mode, "r")

    def test_default_error_propagation_behavior(self):
        self.assertEqual(
            self.hook.propagate_errors, RuntimeAuditHook.PROPAGATE_ERRORS_DEFAULT
        )

    def test_hook_works_if_an_exception_is_raised(self):
        # In the case where an exception is raised in the body of the function, the hook
        # should still work correctly.
        @self.auditor.audit("test.erroneous_func", hooks=[self.hook])
        def erroneous_func(filename):
            with open(filename, "w") as f:
                f.write("Hello!\n")

            # Artificially raise an error at this point
            assert False

        def try_erroneous_func(filename):
            try:
                return erroneous_func(filename)
            except:
                pass

        with tempfile.NamedTemporaryFile() as f:
            try_erroneous_func(f.name)
            with self.auditor.start_auditing():
                try_erroneous_func(f.name)
            try_erroneous_func(f.name)

            self.assertEqual(self.hook.total_file_opens, 1)


class ErroneousRuntimeHookTestCase(RuntimeHookTestCaseMixin, unittest.TestCase):
    """Tests for hook classes that inherit from RuntimeHook that raise an error in
    their sys_hook function."""

    hook_gen = ErroneousHook

    @req_python_version(min=(3, 8))
    def setUp(self):
        super().setUp()

        @self.auditor.audit("my_event", hooks=[self.hook])
        def my_event():
            sys.audit("sys.my_event")

        self.my_event = my_event

    def test_hook_with_no_propagation(self):
        # When error propagation is disabled, errors should instead be logged
        with self.auditor.start_auditing():
            self.my_event()
        output = self.logging_output.getvalue().rstrip()
        output = json.loads(output)
        self.assertEqual(output["level"], "ERROR")
        self.assertEqual(
            output["message"],
            "ValueError raised in ErroneousHook.sys_hook: my_test_message",
        )

    def test_hook_with_propagation(self):
        self.hook.propagate_errors = True
        with self.auditor.start_auditing():
            with self.assertRaises(ValueError):
                self.my_event()


if __name__ == "__main__":
    unittest.main()



import json
import unittest
from collections import Counter
from seagrass import get_current_event, get_audit_logger
from seagrass.base import CleanupHook
from seagrass.hooks import ContextManagerHook
from test.utils import HookTestCaseMixin


class CounterContextManager:
    def __init__(self):
        self.ctr = Counter()
        self.exc_ctr = Counter()

    def __enter__(self):
        self.ctr[get_current_event()] += 1

    def __exit__(self, exc_type, exc_val, traceback):
        if exc_type is not None:
            self.exc_ctr[get_current_event()] += 1


class LoggerContextManager:
    def __enter__(self):
        get_audit_logger().debug(f"Calling {get_current_event()}")

    def __exit__(self, exc_type, exc_val, traceback):
        get_audit_logger().debug(f"Exiting {get_current_event()}")


class ContextManagerHookTestCase(HookTestCaseMixin, unittest.TestCase):

    check_interfaces = (CleanupHook,)

    def setUp(self):
        super().setUp()
        self.cm = CounterContextManager()
        self.hook = ContextManagerHook(self.cm, nest=True)

    def test_hook_function(self):
        @self.auditor.audit("test.foo", hooks=[self.hook])
        def foo():
            pass

        @self.auditor.audit("test.bar", hooks=[self.hook])
        def bar():
            assert False

        with self.auditor.start_auditing():
            foo()
            with self.assertRaises(AssertionError):
                bar()

        self.assertEqual(self.cm.ctr["test.foo"], 1)
        self.assertEqual(self.cm.ctr["test.bar"], 1)
        self.assertEqual(self.cm.exc_ctr["test.foo"], 0)
        self.assertEqual(self.cm.exc_ctr["test.bar"], 1)


class CallableContextManagerHookTestCase(HookTestCaseMixin, unittest.TestCase):
    """Tests for ContextManagerHook in the case where the input context manager is a callable
    function."""

    check_interfaces = (CleanupHook,)

    @staticmethod
    def hook_gen():
        return ContextManagerHook(LoggerContextManager, nest=False)

    def test_hook_function(self):
        @self.auditor.audit("test.foo", hooks=[self.hook])
        def foo():
            pass

        @self.auditor.audit("test.bar", hooks=[self.hook])
        def bar():
            foo()

        with self.auditor.start_auditing():
            bar()

        output = self.logging_output.getvalue().rstrip().split("\n")
        output = [json.loads(o) for o in output]

        self.assertTrue(all(o["level"] == "DEBUG" for o in output))
        self.assertEqual(output[0]["message"], "Calling test.bar")
        self.assertEqual(output[1]["message"], "Exiting test.bar")

        self.hook.nest = True

        with self.auditor.start_auditing():
            bar()

        output = self.logging_output.getvalue().rstrip().split("\n")
        output = output[2:]
        output = [json.loads(o) for o in output]

        self.assertTrue(all(o["level"] == "DEBUG" for o in output))
        self.assertEqual(output[0]["message"], "Calling test.bar")
        self.assertEqual(output[1]["message"], "Calling test.foo")
        self.assertEqual(output[2]["message"], "Exiting test.foo")
        self.assertEqual(output[3]["message"], "Exiting test.bar")


if __name__ == "__main__":
    unittest.main()



# Tests for the LoggerHook auditing hook.

import json
import logging
import unittest
from functools import reduce
from operator import add, mul
from seagrass.hooks import LoggingHook
from test.utils import HookTestCaseMixin, SeagrassTestCaseMixin


class LoggingHookTestCase(HookTestCaseMixin, unittest.TestCase):
    def setUp(self):
        super(HookTestCaseMixin, self).setUp()

        self.hook_pre = LoggingHook(
            prehook_msg=lambda e, args, kwargs: f"hook_pre: {e}, args={args}, kwargs={kwargs}",
        )
        self.hook_both = LoggingHook(
            prehook_msg=lambda e, args, kwargs: f"hook_both: {e}, args={args}, kwargs={kwargs}",
            posthook_msg=lambda e, result: f"hook_both: {e}, result={result}",
            loglevel=logging.INFO,
        )

        # Use hook_both as self.hook for running additional tests that are defined for
        # test case classes that subclass from HookTestMixin.
        self.hook = self.hook_both

    def test_hook_function(self):
        event = "test.multiply_or_add"

        @self.auditor.audit(event, hooks=[self.hook_pre, self.hook_both])
        def multiply_or_add(*args, op="*"):
            if op == "*":
                return reduce(mul, args, 1)
            elif op == "+":
                return reduce(add, args, 0)
            else:
                raise ValueError(f"Unknown operation '{op}'")

        args = (1, 2, 3, 4)
        kwargs_add = {"op": "+"}
        with self.auditor.start_auditing():
            multiply_or_add(*args)
            multiply_or_add(*args, **kwargs_add)

        output = self.logging_output.getvalue().rstrip().split("\n")
        output = [json.loads(o) for o in output]

        self.assertEqual(output[0]["level"], "DEBUG")
        self.assertEqual(
            output[0]["message"], f"hook_pre: {event}, args={args}, kwargs={{}}"
        )
        self.assertEqual(output[1]["level"], "INFO")
        self.assertEqual(
            output[1]["message"], f"hook_both: {event}, args={args}, kwargs={{}}"
        )
        self.assertEqual(output[2]["level"], "INFO")
        self.assertEqual(output[2]["message"], f"hook_both: {event}, result={24}")
        self.assertEqual(output[3]["level"], "DEBUG")
        self.assertEqual(
            output[3]["message"], f"hook_pre: {event}, args={args}, kwargs={kwargs_add}"
        )
        self.assertEqual(output[4]["level"], "INFO")
        self.assertEqual(
            output[4]["message"],
            f"hook_both: {event}, args={args}, kwargs={kwargs_add}",
        )
        self.assertEqual(output[5]["level"], "INFO")
        self.assertEqual(output[5]["message"], f"hook_both: {event}, result={10}")


class MiscellaneousLoggingHookTestCase(SeagrassTestCaseMixin, unittest.TestCase):
    def test_raise_error_if_neither_message_is_specified(self):
        """A ValueError should be raised if we try to create a LoggingHook without providing both
        a prehook_msg and a posthook_msg."""
        with self.assertRaises(ValueError):
            LoggingHook()

    def test_use_alternate_logging_input(self):
        """Construct prehook_msg and posthook_msg such that they provide a tuple of arguments
        rather than an already-formatted string."""

        def prehook_msg(event_name, args, kwargs):
            return ("prehook: event=%s", event_name)

        def posthook_msg(event_name, result):
            return ("posthook: event=%s", event_name)

        hook = LoggingHook(prehook_msg=prehook_msg, posthook_msg=posthook_msg)

        @self.auditor.audit("test.foo", hooks=[hook])
        def foo():
            pass

        with self.auditor.start_auditing():
            foo()

        output = self.logging_output.getvalue().rstrip().split("\n")
        output = [json.loads(o) for o in output]
        self.assertTrue(all(o["level"] == "DEBUG" for o in output))
        self.assertEqual(output[0]["message"], "prehook: event=test.foo")
        self.assertEqual(output[1]["message"], "posthook: event=test.foo")



# Tests for creating events out of async functions

import asyncio
import json
import unittest
from seagrass import get_current_event
from seagrass.hooks import CounterHook, LoggingHook
from test.utils import SeagrassTestCaseMixin, async_test


class AsyncEventTestCase(SeagrassTestCaseMixin, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.ctr_hook = CounterHook()
        self.log_hook = LoggingHook(
            prehook_msg=lambda event, *args: f"Starting {event}",
            posthook_msg=lambda event, *args: f"Leaving {event}",
        )
        self.hooks = [self.ctr_hook, self.log_hook]

    @async_test
    async def test_create_event_over_async_function(self):
        """Ensure that auditing works correctly when we call an async event."""

        @self.auditor.audit("test.foo", use_async=True, hooks=self.hooks)
        async def foo():
            self.assertEqual(get_current_event(), "test.foo")

        with self.auditor.start_auditing():
            await foo()

        self.assertEqual(self.ctr_hook.event_counter["test.foo"], 1)

        output = self.logging_output.getvalue().rstrip().split("\n")
        output = [json.loads(o) for o in output]

        self.assertEqual(len(output), 2)
        self.assertTrue(all(o["level"] == "DEBUG" for o in output))
        self.assertEqual(output[0]["message"], "Starting test.foo")
        self.assertEqual(output[1]["message"], "Leaving test.foo")

    @async_test
    async def test_wrap_nested_async_events(self):
        """Ensure that auditing works correctly when we call an async event inside another
        async event."""

        @self.auditor.async_audit("test.foo", hooks=self.hooks)
        async def foo():
            self.assertEqual(get_current_event(), "test.foo")

        @self.auditor.async_audit("test.bar", hooks=self.hooks)
        async def bar(recurse: bool):
            self.assertEqual(get_current_event(), "test.bar")
            await foo()
            self.assertEqual(get_current_event(), "test.bar")

            if recurse:
                await asyncio.gather(bar(False), foo(), bar(False))

            self.assertEqual(get_current_event(), "test.bar")

        with self.auditor.start_auditing():
            await bar(True)

        self.assertEqual(self.ctr_hook.event_counter["test.foo"], 4)
        self.assertEqual(self.ctr_hook.event_counter["test.bar"], 3)

    @async_test
    async def test_wrapped_async_and_sync_events(self):
        """Check that auditing works correctly if we run a sync event inside an async event."""

        @self.auditor.audit("test.foo", hooks=self.hooks)
        def foo():
            self.assertEqual(get_current_event(), "test.foo")

        @self.auditor.async_audit("test.bar", hooks=self.hooks)
        async def bar(recurse: bool):
            self.assertEqual(get_current_event(), "test.bar")
            foo()
            self.assertEqual(get_current_event(), "test.bar")

            if recurse:
                await bar(False)
                foo()
                await bar(False)

            self.assertEqual(get_current_event(), "test.bar")

        with self.auditor.start_auditing():
            await bar(True)

        self.assertEqual(self.ctr_hook.event_counter["test.foo"], 4)
        self.assertEqual(self.ctr_hook.event_counter["test.bar"], 3)


if __name__ == "__main__":
    unittest.main()



import seagrass
import seagrass._typing as t
import unittest
from seagrass import get_audit_logger
from seagrass.base import ProtoHook, CleanupHook
from seagrass.hooks import TracingHook
from test.utils import HookTestCaseMixin
from types import FrameType


class EmptyTracingHook(TracingHook):
    """Example TracingHook where the tracing function does nothing."""

    def __init__(self) -> None:
        return super().__init__(self.tracefunc)

    def tracefunc(
        self, frame: FrameType, event: str, arg: t.Any
    ) -> t.Optional[TracingHook.TraceFunc]:
        return self.tracefunc


class LocalVariableExtractorHook(TracingHook):
    """Example TracingHook that extracts the value of MY_TEST_VARIABLE from the current frame's
    'locals' dictionary. The hook always stores the last value of MY_TEST_VARIABLE that it saw.
    """

    def __init__(self):
        super().__init__(self.tracefunc)
        self.reset()

    def tracefunc(
        self, frame: FrameType, event: str, arg: t.Any
    ) -> t.Optional[TracingHook.TraceFunc]:
        if "MY_TEST_VARIABLE" in frame.f_locals:
            self.last_event = self.current_event
            self.MY_TEST_VARIABLE = frame.f_locals["MY_TEST_VARIABLE"]
        return self.tracefunc

    def reset(self):
        self.last_event = None
        self.MY_TEST_VARIABLE = None


class TracingHookTestCase(HookTestCaseMixin, unittest.TestCase):

    check_interfaces = (CleanupHook,)

    @staticmethod
    def hook_gen():
        return LocalVariableExtractorHook()

    def test_hook_function(self):
        """Hook a function using a TracingHook."""

        @self.auditor.audit("event.foo", hooks=[self.hook])
        def foo(x):
            MY_TEST_VARIABLE = x
            get_audit_logger().info(f"MY_TEST_VARIABLE={MY_TEST_VARIABLE}")

        @self.auditor.audit("event.bar", hooks=[self.hook])
        def bar():
            MY_TEST_VARIABLE = 1337
            get_audit_logger().info(f"MY_TEST_VARIABLE={MY_TEST_VARIABLE}")

        with self.auditor.start_auditing(reset_hooks=True):
            foo(42)
            self.assertEqual(self.hook.MY_TEST_VARIABLE, 42)
            self.assertEqual(self.hook.last_event, "event.foo")

            bar()
            self.assertEqual(self.hook.MY_TEST_VARIABLE, 1337)
            self.assertEqual(self.hook.last_event, "event.bar")

            # Outside of events, the is_active property should be False
            self.assertEqual(self.hook.is_active, False)

    def test_nest_tracing_hook_events(self):
        """Nest multiple events using the same TracingHook."""

        # Create a test hook we'll use to perform tests after each event
        assertEqual = self.assertEqual
        trace_hook = self.hook

        class TestHook(ProtoHook[None]):
            def prehook(self, event, args, kwargs):
                if event == "event.foo":
                    assertEqual(trace_hook.MY_TEST_VARIABLE, 42 + 1)

            def posthook(self, event, result, context):
                if event == "event.foo":
                    assertEqual(trace_hook.MY_TEST_VARIABLE, 2 * (42 + 1))
                if event == "event.bar":
                    assertEqual(trace_hook.MY_TEST_VARIABLE, 2 * (42 + 1) - 5)

        test_hook = TestHook()

        @self.auditor.audit("event.foo", hooks=[trace_hook, test_hook])
        def foo(x):
            self.assertTrue(trace_hook.is_active)
            MY_TEST_VARIABLE = 2 * x
            return MY_TEST_VARIABLE

        @self.auditor.audit("event.bar", hooks=[trace_hook, test_hook])
        def bar(x):
            MY_TEST_VARIABLE = x + 1
            MY_TEST_VARIABLE = foo(MY_TEST_VARIABLE) - 5
            get_audit_logger().info(f"MY_TEST_VARIABLE={MY_TEST_VARIABLE}")
            self.assertTrue(trace_hook.is_active)

        with self.auditor.start_auditing(reset_hooks=True):
            bar(42)

    def test_cannot_instantiate_more_than_one_tracing_hook(self):
        """The trace functions set by multiple TracingHooks can override one another. As such,
        an error should be raised whenever we try to active more than one TracingHook."""

        variable_hook = LocalVariableExtractorHook()
        empty_hook = EmptyTracingHook()

        # Of the three functions defined below, "foo" and "bar" should work fine, since only
        # one TracingHook is activated for each of them. "baz" should error out since it
        # causes both of the TracingHooks to be activated.
        @self.auditor.audit(seagrass.auto, hooks=[empty_hook])
        def foo():
            self.assertTrue(empty_hook.is_active)

        @self.auditor.audit(seagrass.auto, hooks=[variable_hook])
        def bar():
            self.assertTrue(variable_hook.is_active)

        @self.auditor.audit(seagrass.auto, hooks=[empty_hook, variable_hook])
        def baz():
            pass

        with self.auditor.start_auditing():
            foo()
            bar()
            with self.assertRaises(ValueError):
                baz()
            foo()
            bar()


if __name__ == "__main__":
    unittest.main()



# flake8: noqa: F401

# Type annotations for Seagrass
# Seagrass code should import type annotations from this module rather
# than from `typing` to ensure version compatibility

from typing import *

# The __all__ list includes all of the attributes of typing that were
# brought into scope.
import typing as _t

__all__: List[str] = []
__all__ += _t.__all__  # type: ignore[attr-defined]

# Import additional attributes for Python < 3.8
_extended_attrs = ["Final", "Literal", "Protocol", "runtime_checkable"]

# Additional types for Seagrass:
class Missing:
    """Unique type used throughout Seagrass to represent a missing value."""

    __slots__: List[str] = []

    def __repr__(self) -> str:
        return f"<{__name__}.{self.__class__.__name__}>"


MISSING: Final[Missing] = Missing()

_T = TypeVar("_T")

# Maybe[T] is a type that represents a value that is potentially missing its value.
# This is distinct from Optional[T], which represents a value that could have type
# T or that could be None. In cases where a None value should be allowed, this type
# may be used instead.
Maybe = Union[_T, Missing]

_R = TypeVar("_R")
_F = TypeVar("_F", bound=Callable)
_FR = TypeVar("_FR", bound=Callable[..., _R])


class AuditedFunc(Protocol[_FR]):

    __event_name__: str

    def __call__(*args, **kwargs) -> _R:    # type: ignore[type-var]
        ...


class AuditDecorator(Protocol):
    def __call__(self, func: _F) -> _F:
        ...


# Add additional types to the __all__ list:

__all__ += [
    "Missing",
    "MISSING",
    "Maybe",
    "AuditedFunc",
    "AuditDecorator",
]

__all__ = sorted(__all__)



import logging
import seagrass._typing as _t


class SeagrassLogFilter(logging.Filter):
    """A custom :py:class:`logging.Filter` that attaches contextual information
    about Seagrass events to logs."""

    def filter(self, record) -> bool:
        from seagrass.events import get_current_event
        from seagrass.hooks import get_hook_context

        seagrass_context: _t.Dict[str, _t.Any] = {}
        seagrass_context["event"] = get_current_event(default=None)

        if (hook_ctx := get_hook_context(default=None)) is not None:
            seagrass_context["hook"] = hook_ctx.name
            seagrass_context["hook_ctx"] = hook_ctx.args

        if hasattr(record, "seagrass"):
            record.seagrass = {**record.seagrass, **seagrass_context}
        else:
            record.seagrass = seagrass_context

        return True



import seagrass._typing as t
from seagrass.base import ProtoHook
from types import TracebackType

C = t.TypeVar("C")
CtxType = t.Optional[t.ContextManager[C]]


class ContextManagerHook(ProtoHook[CtxType[C]]):
    """A hook that wraps around a context manager, calling the context manager's ``__enter__`` and
    ``__exit__`` methods whenever the hook is activated."""

    __current_cm: t.Optional[t.ContextManager[C]] = None

    @property
    def current_context_manager(self) -> t.Optional[t.ContextManager[C]]:
        """Return the current context manager instance being used by the hook."""
        return self.__current_cm

    @property
    def is_active(self) -> bool:
        """Returns ``True`` if an event that uses this hook is currently being executed."""
        return self.current_context_manager is not None

    def __init__(
        self,
        cm: t.Union[t.ContextManager[C], t.Callable[[], t.ContextManager[C]]],
        nest: bool = False,
    ) -> None:
        """Create a new :py:class:`ContextManagerHook`.

        :param t.Union[t.ContextManager, t.Callable[[], t.ContextManager]] cm: a context manager
            (i.e., an object with ``__enter__`` and ``__exit__`` methods) or a function that takes
            no arguments and creates a new context manager.
        :param bool nest: whether successive invocations of the context manager should be nested.
            If ``nest=False`` and another Seagrass event that uses this hook gets called, the
            prehook and cleanup stages of the hook will be skipped. In general, you should set
            ``nest=True`` if your context manager should be used along the lines of

                .. code::

                    with cm:
                        # Execute outer event
                        ...
                        with cm:
                            # Execute inner event
                            ...
                        # Continue executing outer event
                        ...

            and ``nest=False`` if your context manager should be used as

                .. code::

                    with cm:
                        # Execute outer event
                        ...
                        # Execute inner event
                        ...
                        # Continue executing outer event
                        ...
        """
        self.cm = cm
        self.nest = nest

    def __create_cm(self) -> t.ContextManager[C]:
        if isinstance(self.cm, t.ContextManager):
            return self.cm
        else:
            return self.cm()

    def prehook(
        self, event: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> CtxType[C]:
        if self.nest or self.current_context_manager is None:
            cm = self.__create_cm()
            cm.__enter__()
        else:
            cm = self.current_context_manager

        old_cm = self.current_context_manager
        self.__current_cm = cm
        return old_cm

    def cleanup(
        self,
        event: str,
        context: CtxType[C],
        exc: t.Tuple[
            t.Optional[t.Type[BaseException]],
            t.Optional[BaseException],
            t.Optional[TracebackType],
        ],
    ) -> None:
        old_cm = context
        current_cm = self.current_context_manager
        self.__current_cm = old_cm

        if (self.nest or not self.is_active) and current_cm is not None:
            exc_type, exc_val, tb = exc
            current_cm.__exit__(exc_type, exc_val, tb)



# Tests for the TimerHook auditing hook

import json
import time
import unittest
from test.utils import HookTestCaseMixin
from seagrass.base import LogResultsHook, ResettableHook
from seagrass.hooks import TimerHook


class TimerHookTestCase(HookTestCaseMixin, unittest.TestCase):

    hook_gen = TimerHook
    check_interfaces = (LogResultsHook, ResettableHook)

    def test_hook_function(self):
        ausleep = self.auditor.audit("test.time.sleep", time.sleep, hooks=[self.hook])

        ausleep(0.01)
        with self.auditor.start_auditing():
            ausleep(0.01)
        ausleep(0.01)

        recorded_time = self.hook.event_times["test.time.sleep"]
        self.assertAlmostEqual(recorded_time, 0.01, delta=0.005)

        # Check logging output
        self.auditor.log_results()
        self.logging_output.seek(0)
        output = [line.rstrip() for line in self.logging_output.readlines()]
        output = [json.loads(line) for line in output]

        self.assertTrue(all(o.get("level") == "INFO" for o in output))
        self.assertEqual(len(output), 1)
        self.assertEqual(output[0]["message"], "TimerHook results")
        self.assertEqual(output[0]["seagrass"]["hook"], "TimerHook")
        self.assertEqual(output[0]["seagrass"]["hook_ctx"]["event"], "test.time.sleep")
        self.assertEqual(output[0]["seagrass"]["hook_ctx"]["time"], recorded_time)


if __name__ == "__main__":
    unittest.main()



# Tests for the CounterHook auditing hook.

from seagrass.base import ResettableHook, LogResultsHook
from seagrass.hooks import CounterHook
from test.utils import HookTestCaseMixin
import json
import unittest


class CounterHookTestCase(HookTestCaseMixin, unittest.TestCase):

    hook_gen = CounterHook
    check_interfaces = (ResettableHook, LogResultsHook)

    def test_hook_function(self):
        @self.auditor.audit("test.say_hello", hooks=[self.hook])
        def say_hello(name: str) -> str:
            return f"Hello, {name}!"

        self.assertEqual(self.hook.event_counter["test.say_hello"], 0)

        # Hook should not get called outside of an auditing context
        say_hello("Alice")
        self.assertEqual(self.hook.event_counter["test.say_hello"], 0)

        with self.auditor.start_auditing():
            for name in ("Alice", "Bob", "Cathy"):
                say_hello(name)
        self.assertEqual(self.hook.event_counter["test.say_hello"], 3)
        self.assertEqual(set(self.hook.event_counter), set(("test.say_hello",)))

        # Upon resetting the hook, all event counts should be set back to zero
        self.hook.reset()
        self.assertEqual(self.hook.event_counter["test.say_hello"], 0)

    def test_counter_hook_logging(self):
        # Collect event information with CounterHook and then check the logs
        # that are emitted by CounterHook.log_results.

        self.auditor.create_event("event_b", hooks=[self.hook])
        self.auditor.create_event("event_a", hooks=[self.hook])
        self.auditor.create_event("event_c", hooks=[self.hook])

        with self.auditor.start_auditing(log_results=True):
            for _ in range(904):
                self.auditor.raise_event("event_b")
            for _ in range(441):
                self.auditor.raise_event("event_a")
            for _ in range(58):
                self.auditor.raise_event("event_c")

        lines = self.logging_output.getvalue().rstrip().split("\n")
        output = [json.loads(line) for line in lines]
        self.assertEqual(len(lines), 3)
        self.assertTrue(all(o["level"] == "INFO" for o in output))
        self.assertTrue(all(o["seagrass"]["hook"] == "CounterHook" for o in output))

        self.assertEqual(output[0]["seagrass"]["hook_ctx"]["event"], "event_a")
        self.assertEqual(output[0]["seagrass"]["hook_ctx"]["count"], 441)
        self.assertEqual(output[1]["seagrass"]["hook_ctx"]["event"], "event_b")
        self.assertEqual(output[1]["seagrass"]["hook_ctx"]["count"], 904)
        self.assertEqual(output[2]["seagrass"]["hook_ctx"]["event"], "event_c")
        self.assertEqual(output[2]["seagrass"]["hook_ctx"]["count"], 58)

    def test_disable_counter_hook(self):
        self.auditor.create_event("my_event", hooks=[self.hook])

        with self.auditor.start_auditing(reset_hooks=True):
            self.auditor.raise_event("my_event")
            self.assertEqual(self.hook.event_counter["my_event"], 1)

        self.hook.enabled = False

        with self.auditor.start_auditing(reset_hooks=True):
            self.auditor.raise_event("my_event")
            self.assertEqual(self.hook.event_counter["my_event"], 0)


if __name__ == "__main__":
    unittest.main()



# This module primarily exists for the purposes of helping generate the
# Seagrass documentation. The code in this module is not part of the
# public Seagrass API.

import logging
import logging.config
import sys
from pythonjsonlogger import jsonlogger     # type: ignore
from seagrass import DEFAULT_LOGGER_NAME


class LogFormatter(jsonlogger.JsonFormatter):   # type: ignore
    def add_fields(self, log_record, record, message_dict):
        super(LogFormatter, self).add_fields(log_record, record, message_dict)
        log_record["level"] = record.levelname


def configure_logging(name: str = DEFAULT_LOGGER_NAME) -> logging.Logger:
    """Set up the default logging configuration for the documentation."""

    logger = logging.getLogger(name)
    while logger.hasHandlers():
        logger.removeHandler(logger.handlers[0])

    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = LogFormatter()
    handler.setFormatter(formatter)

    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)

    return logger



# Tests for protocols and functions defined in seagrass.base

import seagrass.base as base
import seagrass._typing as t
import unittest


class CustomHookImplementationTestCase(unittest.TestCase):
    def setUp(self):
        # Create a version of the ProtoHook protocol that we can check at runtime.
        self.CheckableProtoHook = t.runtime_checkable(base.ProtoHook)

    def test_get_hook_priority(self):
        class MyHook(base.ProtoHook[None]):
            def prehook(self, *args):
                ...

            def posthook(self, *args):
                ...

        hook = MyHook()
        self.assertIsInstance(
            hook,
            self.CheckableProtoHook,
            f"{hook.__class__.__name__} does not satisfy the hooking protocol",
        )
        self.assertEqual(hook.priority, base.DEFAULT_PRIORITY)


if __name__ == "__main__":
    unittest.main()



# flake8: noqa: F401

from .context_manager_hook import ContextManagerHook
from .counter_hook import CounterHook
from .file_open_hook import FileOpenHook
from .logging_hook import LoggingHook
from .profiler_hook import ProfilerHook
from .runtime_audit_hook import RuntimeAuditHook
from .stack_trace_hook import StackTraceHook
from .timer_hook import TimerHook
from .tracing_hook import TracingHook

__all__ = [
    "ContextManagerHook",
    "CounterHook",
    "FileOpenHook",
    "LoggingHook",
    "ProfilerHook",
    "StackTraceHook",
    "RuntimeAuditHook",
    "TimerHook",
    "TracingHook",
]

from ._utils import get_hook_context, HookContext

__all__ += ["get_hook_context", "HookContext"]



import seagrass._typing as t
from contextvars import ContextVar
from dataclasses import dataclass
from seagrass.base import ProtoHook, CleanupHook
from types import TracebackType

T = t.TypeVar("T")

# Context variable used to store the current event
current_event: ContextVar[str] = ContextVar("seagrass_current_event")


@t.overload
def get_current_event() -> str:
    ...  # pragma: no cover


@t.overload
def get_current_event(default: T) -> t.Union[str, T]:
    ...  # pragma: no cover


def get_current_event(default: t.Maybe[T] = t.MISSING) -> t.Union[str, T]:
    """Get the current Seagrass event that is being executed.

    :raises LookupError: if no Seagrass event is currently under execution, and
        ``default`` is not specified.
    """
    if isinstance(default, t.Missing):
        return current_event.get()
    else:
        return current_event.get(default)


@dataclass
class EventData:
    """A class that holds data corresponding to the event being executed."""

    event_name: str
    args: t.Tuple[t.Any, ...]
    kwargs: t.Dict[str, t.Any]
    result: t.Maybe[t.Any] = t.MISSING


class HookExecutionContext(t.Generic[T]):
    """A context manager that wraps around a hook that is in charge of its prehook and posthook."""

    __slots__: t.List[str] = ["data", "hook", "prehook_context"]

    def __init__(self, hook: ProtoHook[T], data: EventData) -> None:
        self.data = data
        self.hook = hook

    def __enter__(self) -> "HookExecutionContext":
        self.prehook_context: T = self.hook.prehook(
            self.data.event_name, self.data.args, self.data.kwargs
        )
        return self

    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        tb: t.Optional[TracebackType],
    ) -> None:
        try:
            if not isinstance(self.data.result, t.Missing):
                self.hook.posthook(
                    self.data.event_name, self.data.result, self.prehook_context
                )
        finally:
            if isinstance(self.hook, CleanupHook):
                exc = (exc_type, exc_val, tb)
                self.hook.cleanup(self.data.event_name, self.prehook_context, exc)

    async def __aenter__(self) -> "HookExecutionContext":
        return self.__enter__()

    async def __aexit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        tb: t.Optional[TracebackType],
    ) -> None:
        return self.__exit__(exc_type, exc_val, tb)



import seagrass
import sys
import unittest
import warnings
from collections import Counter, defaultdict
from seagrass import auto, get_current_event
from seagrass.hooks import CounterHook
from test.utils import SeagrassTestCaseMixin, req_python_version

with seagrass.create_global_auditor() as _:

    class ExampleClass:
        # Test class used to check how functions are auto-named by Seagrass
        @staticmethod
        @seagrass.audit(auto)
        def say_hello(name: str) -> str:
            return f"Hello, {name}!"


class EventsTestCase(SeagrassTestCaseMixin, unittest.TestCase):
    """Tests for events created by Seagrass."""

    def test_wrap_class_property(self):
        # Override a class property to call a hook whenever it's accessed
        class Foo:
            def __init__(self):
                self.x = 0

            def add_one(self):
                return self.x + 1

        hook = CounterHook()

        @self.auditor.audit("test.foo.get_x", hooks=[hook])
        def get_x(self):
            return self.__x

        @self.auditor.audit("test.foo.set_x", hooks=[hook])
        def set_x(self, val):
            self.__x = val

        @self.auditor.audit("test.foo.del_x", hooks=[hook])
        def del_x(self):
            del self.__x

        setattr(Foo, "x", property(fget=get_x, fset=set_x, fdel=del_x))

        with self.auditor.start_auditing():
            f = Foo()
            f.x = 1
            y = f.x  # noqa: F841
            f.x += 2
            del f.x

        # We call get_x twice (once for y = f.x, another for f.x += 2)
        # We call set_x three times (once during Foo.__init__, once during f.x = 1, and
        #   once during f.x += 2)
        # We call del_x once (when we call del f.x)
        self.assertEqual(hook.event_counter["test.foo.get_x"], 2)
        self.assertEqual(hook.event_counter["test.foo.set_x"], 3)
        self.assertEqual(hook.event_counter["test.foo.del_x"], 1)

        # Now override the add_one function belonging to Foo
        current_add_one = Foo.add_one

        @self.auditor.audit("test.foo.add_one", hooks=[hook])
        def add_one(self, *args, **kwargs):
            return current_add_one(self, *args, **kwargs)

        setattr(Foo, "add_one", add_one)

        with self.auditor.start_auditing():
            f = Foo()
            result = f.add_one()

        self.assertEqual(result, 1)
        self.assertEqual(hook.event_counter["test.foo.add_one"], 1)

    def test_toggle_event(self):
        hook = CounterHook()

        @self.auditor.audit("test.foo", hooks=[hook])
        def foo():
            return

        @self.auditor.audit("test.bar", hooks=[hook])
        def bar():
            return foo()

        with self.auditor.start_auditing():
            bar()
            self.assertEqual(hook.event_counter["test.foo"], 1)
            self.assertEqual(hook.event_counter["test.bar"], 1)

            # After disabling an event, its event hooks should no longer be called
            self.auditor.toggle_event("test.foo", False)
            bar()
            self.assertEqual(hook.event_counter["test.foo"], 1)
            self.assertEqual(hook.event_counter["test.bar"], 2)

            # Now we re-enable the event so that hooks get called again
            self.auditor.toggle_event("test.foo", True)
            bar()
            self.assertEqual(hook.event_counter["test.foo"], 2)
            self.assertEqual(hook.event_counter["test.bar"], 3)

    @req_python_version(min=(3, 8))
    def test_wrap_function_and_create_sys_audit_event(self):
        # We should be able to set up sys.audit events when we wrap functions
        @self.auditor.audit("test.foo", raise_runtime_events=True)
        def foo(x, y, z=None):
            return x + y + (0 if z is None else z)

        @self.auditor.audit("test.bar", raise_runtime_events=False)
        def bar(x, y, z=None):
            return x + y + (0 if z is None else z)

        @self.auditor.audit(
            "test.baz",
            raise_runtime_events=True,
            prehook_audit_event_name="baz_prehook",
            posthook_audit_event_name="baz_posthook",
        )
        def baz(x, y, z=None):
            return x + y + (0 if z is None else z)

        events_counter = Counter()
        args_dict = defaultdict(list)

        def audit_hook(event: str, *args):
            try:
                if event.startswith("prehook:") or event.startswith("posthook:"):
                    events_counter[event] += 1
                    args_dict[event].append(args)
                elif event in ("baz_prehook", "baz_posthook"):
                    events_counter[event] += 1
                    args_dict[event].append(args)
            except Exception as ex:
                warnings.warn(f"Exception raised in audit_hook: ex={ex}")

        sys.addaudithook(audit_hook)

        test_args = [(-3, 4), (5, 8), (0, 0)]
        test_kwargs = [{}, {}, {"z": 1}]

        def run_fns(args_list, kwargs_list):
            for (args, kwargs) in zip(args_list, kwargs_list):
                for fn in (foo, bar, baz):
                    fn(*args, **kwargs)

        # The following call to run_fns shouldn't raise any audit events since
        # it isn't performed in an auditing context.
        run_fns(test_args, test_kwargs)
        self.assertEqual(set(events_counter), set())
        self.assertEqual(set(args_dict), set())

        # Now some audit events should be raised:
        with self.auditor.start_auditing():
            run_fns(test_args, test_kwargs)

        expected_prehooks = ["prehook:test.foo", "baz_prehook"]
        expected_posthooks = ["posthook:test.foo", "baz_posthook"]
        self.assertEqual(
            set(events_counter), set(expected_prehooks + expected_posthooks)
        )
        self.assertEqual(set(events_counter), set(args_dict))

        for event in expected_prehooks:
            self.assertEqual(events_counter[event], len(test_args))
            args = [args[0][0] for args in args_dict[event]]
            kwargs = [args[0][1] for args in args_dict[event]]
            self.assertEqual(args, test_args)
            self.assertEqual(kwargs, test_kwargs)

        # If we try running our functions outside of an auditing context again,
        # we should once again find that no system events are raised.
        events_counter.clear()
        args_dict.clear()
        run_fns(test_args, test_kwargs)
        self.assertEqual(set(events_counter), set())
        self.assertEqual(set(args_dict), set())

    def test_auto_name_event(self):
        from pathlib import Path

        auhome = self.auditor.audit(auto, Path.home)
        self.assertEqual(auhome.__event_name__, "pathlib.Path.home")

        # Check the name of the audited function from the ExampleClass class at
        # the top of the file.
        self.assertEqual(
            ExampleClass.say_hello.__event_name__, f"{__name__}.ExampleClass.say_hello"
        )

    def test_get_current_event(self):
        @self.auditor.audit("test.foo")
        def foo():
            self.assertEqual(get_current_event(), "test.foo")

        @self.auditor.audit("test.bar")
        def bar():
            self.assertEqual(get_current_event(), "test.bar")
            foo()
            self.assertEqual(get_current_event(), "test.bar")

        with self.auditor.start_auditing():
            foo()
            bar()

        # We should be able to specify a default value for get_current_event(). If no default is
        # specified and an event isn't being executed, an exception should be thrown.
        self.assertEqual(get_current_event(None), None)
        with self.assertRaises(LookupError):
            get_current_event()


if __name__ == "__main__":
    unittest.main()
